# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""lintian workflow."""

from collections.abc import Callable
from contextlib import suppress
from typing import Any, override

from debian.debian_support import Version
from django.utils import timezone

from debusine.artifacts import LintianArtifact, SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
    get_source_package_version,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, ArtifactRelation, WorkRequest
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.models import (
    LintianWorkflowData,
    LintianWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.regression_tracking import (
    RegressionTrackingWorkflow,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LintianData,
    LintianFailOnSeverity,
    LintianInput,
    LintianOutput,
    LookupMultiple,
    LookupSingle,
    OutputData,
    RegressionAnalysis,
    RegressionAnalysisStatus,
)


class LintianWorkflow(
    RegressionTrackingWorkflow[LintianWorkflowData, LintianWorkflowDynamicData]
):
    """Lintian workflow."""

    TASK_NAME = "lintian"

    def _has_current_reference_qa_result(
        self, *, architecture: str, build_architecture: str
    ) -> bool:
        """
        Return True iff we have a current reference QA result.

        A lintian analysis is outdated if:

        * either the underlying source or binary packages are outdated (i.e.
          have different version numbers) compared to what's available in
          the ``debian:suite`` collection
        * or the lintian version used to perform the analysis is older than
          the version available in the ``debian:suite`` collection

        Otherwise, it is current.
        """
        # This method is only called when update_qa_results is "yes", in
        # which case these are checked by a model validator.
        assert self.qa_suite is not None
        assert self.reference_qa_results is not None

        source_data = workflow_utils.source_package_data(self)
        latest_result = self.reference_qa_results.manager.lookup(
            f"latest:lintian_{source_data.name}_{architecture}"
        )
        reference_lintian = self.qa_suite.manager.lookup(
            f"binary:lintian_{build_architecture}"
        )
        return (
            latest_result is not None
            and latest_result.artifact is not None
            and latest_result.data["version"] == source_data.version
            and (
                reference_lintian is None
                or Version(
                    LintianArtifact.create_data(
                        latest_result.artifact.data
                    ).summary.lintian_version
                )
                >= Version(reference_lintian.data["version"])
            )
        )

    @override
    def populate(self) -> None:
        """Create work requests."""
        architectures = workflow_utils.get_lookup_architectures(
            self, self.data.binary_artifacts
        )

        if (data_archs := self.data.architectures) is not None:
            architectures.intersection_update(data_archs)

        if architectures != {"all"}:
            architectures = architectures - {"all"}

        if not architectures and not self.data.binary_artifacts:
            # There are no architectures to run on, but that's because the
            # workflow was started with an empty `binary_artifacts`.  In
            # that case there's no point waiting for binary packages to be
            # built, so we might as well just analyze the source package.
            architectures = {"all"}

        # Pick a preferred architecture to produce the common source and
        # binary-all analysis artifacts, to avoid redundancy.  Note that we
        # still pass source and binary-all artifacts to all child work
        # requests so that `lintian` has the best tag coverage available.
        source_all_analysis_architecture: str | None = None
        if "all" in architectures:
            source_all_analysis_architecture = "all"
        elif self.data.arch_all_build_architecture in architectures:
            source_all_analysis_architecture = (
                self.data.arch_all_build_architecture
            )
        elif architectures:
            # Alphabetical sorting doesn't necessarily produce good results
            # here, but this is already a last-ditch fallback case so we
            # don't worry about it too much.  If it turns out to be wrong,
            # it's probably best to just set `arch_all_build_architecture`.
            source_all_analysis_architecture = sorted(architectures)[0]

        environment = f"{self.data.vendor}/match:codename={self.data.codename}"

        for arch in sorted(architectures):
            output = self.data.output.model_copy()
            if arch != source_all_analysis_architecture:
                output.source_analysis = False
                output.binary_all_analysis = False

            if (
                not output.source_analysis
                and not output.binary_all_analysis
                and not output.binary_any_analysis
            ):
                continue

            source_artifact = (
                workflow_utils.locate_debian_source_package_lookup(
                    self, "source_artifact", self.data.source_artifact
                )
            )
            filtered_binary_artifacts = (
                workflow_utils.filter_artifact_lookup_by_arch(
                    self, self.data.binary_artifacts, {arch, "all"}
                )
            )

            self._populate_lintian(
                source_artifact=source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                output=output,
                build_architecture=(
                    self.data.arch_all_build_architecture
                    if arch == "all"
                    else arch
                ),
                environment=environment,
                backend=self.data.backend,
                architecture=arch,
                include_tags=self.data.include_tags,
                exclude_tags=self.data.exclude_tags,
                fail_on_severity=self.data.fail_on_severity,
                target_distribution=f"{self.data.vendor}:{self.data.codename}",
            )

        self._populate_regression_analysis_callback(architectures)

    def _populate_lintian(
        self,
        *,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        output: LintianOutput,
        build_architecture: str,
        environment: str,
        backend: BackendType,
        include_tags: list[str],
        exclude_tags: list[str],
        fail_on_severity: LintianFailOnSeverity,
        architecture: str,
        target_distribution: str,
    ) -> None:
        """Create work request for Lintian for a specific architecture."""
        if (
            self.data.update_qa_results == RegressionTrackingUpdateQAResults.YES
            and self._has_current_reference_qa_result(
                architecture=architecture, build_architecture=build_architecture
            )
        ):
            return

        workflow_data_step = f"lintian-{architecture}"
        workflow_data_kwargs: dict[str, Any] = {}
        if self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO:
            # When updating reference results for regression tracking, task
            # failures never cause the parent workflow or dependent tasks to
            # fail.
            workflow_data_kwargs["allow_failure"] = True

        display_name = f"Lintian for {architecture}"
        if architecture != "all" and output.binary_all_analysis:
            display_name += "+all"

        if output.source_analysis:
            display_name += "+source"

        wr, created = self.work_request_ensure_child_worker(
            task_name="lintian",
            task_data=LintianData(
                input=LintianInput(
                    source_artifact=source_artifact,
                    binary_artifacts=binary_artifacts,
                ),
                output=output,
                build_architecture=build_architecture,
                environment=environment,
                backend=backend,
                include_tags=include_tags,
                exclude_tags=exclude_tags,
                fail_on_severity=fail_on_severity,
                target_distribution=target_distribution,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name=display_name,
                step=workflow_data_step,
                **workflow_data_kwargs,
            ),
        )
        self.requires_artifact(wr, source_artifact)
        self.requires_artifact(wr, binary_artifacts)
        promise_name = f"{self.data.prefix}lintian-{architecture}"
        self.provides_artifact(wr, ArtifactCategory.LINTIAN, promise_name)

        if created and (
            self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO
        ):
            # Checked by a model validator.
            assert self.data.reference_qa_results is not None

            source_data = workflow_utils.source_package_data(self)

            # Back off if another workflow gets there first.
            self.skip_if_qa_result_changed(
                wr,
                package=source_data.name,
                architecture=architecture,
                promise_name=promise_name,
            )

            # Record results in the reference collection.
            action = ActionUpdateCollectionWithArtifacts(
                collection=self.data.reference_qa_results,
                variables={
                    "package": source_data.name,
                    "version": source_data.version,
                    # While this field is technically optional at the
                    # moment, it will be present in any newly-created
                    # artifacts.
                    "$architecture": "architecture",
                    "timestamp": int(
                        (self.qa_suite_changed or timezone.now()).timestamp()
                    ),
                    "work_request_id": wr.id,
                },
                artifact_filters={"category": ArtifactCategory.LINTIAN},
            )
            wr.add_event_reaction("on_success", action)
            wr.add_event_reaction("on_failure", action)

    def _populate_regression_analysis_callback(
        self, architectures: set[str]
    ) -> None:
        """Create a callback to collate regression analysis results."""
        if self.data.enable_regression_tracking:
            # Checked by a model validator.
            assert self.data.reference_prefix

            regression_analysis, _ = self.work_request_ensure_child_internal(
                task_name="workflow",
                workflow_data=WorkRequestWorkflowData(
                    allow_dependency_failures=True,
                    step="regression-analysis",
                    display_name="Regression analysis",
                    visible=False,
                ),
                unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
            )
            for architecture in sorted(architectures):
                for prefix in (self.data.reference_prefix, self.data.prefix):
                    with suppress(KeyError):
                        self.requires_artifact(
                            regression_analysis,
                            f"internal@collections/name:"
                            f"{prefix}lintian-{architecture}",
                        )
            self.orchestrate_child(regression_analysis)

    @override
    def compute_dynamic_data(self) -> LintianWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :parameter_summary: package name_version
        :source_artifact_id: id of the source artifact
        :binary_artifacts_ids: ids of the binary artifacts
        """
        source_artifact = workflow_utils.source_package(self)
        source_data = SourcePackage.create_data(source_artifact.data)

        binary_artifacts_results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        return LintianWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            source_artifact_id=source_artifact.id,
            binary_artifacts_ids=artifact_ids(binary_artifacts_results),
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return source_artifact and the binary artifacts."""
        return [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]

    def _extract_artifact_details(
        self, artifact: Artifact | None
    ) -> tuple[str | None, int | None, WorkRequest | None]:
        """Extract details from a ``debian:lintian`` artifact."""
        source_version: str | None
        artifact_id: int | None
        wr: WorkRequest | None
        if artifact is None:
            source_version = None
            artifact_id = None
            wr = None
        else:
            analyzed_artifact = (
                artifact.relations.filter(
                    type=ArtifactRelation.Relations.RELATES_TO
                )
                .earliest("id")
                .target
            )
            analyzed_data = analyzed_artifact.create_data()
            assert isinstance(
                analyzed_data,
                (
                    DebianSourcePackage,
                    DebianBinaryPackage,
                    DebianUpload,
                ),
            )
            source_version = get_source_package_version(analyzed_data)
            artifact_id = artifact.id
            wr = artifact.created_by_work_request
        return source_version, artifact_id, wr

    @staticmethod
    def compare_qa_results_fine_grained(
        reference: Artifact | None, new: Artifact | None
    ) -> tuple[RegressionAnalysisStatus, dict[str, list[str]]]:
        """Do a fine-grained comparison of two QA results."""
        reference_summary = (
            {} if reference is None else reference.data["summary"]
        )
        reference_count = reference_summary.get("tags_count_by_severity", {})
        reference_found = reference_summary.get("tags_found", [])
        new_summary = {} if new is None else new.data["summary"]
        new_count = new_summary.get("tags_count_by_severity", {})
        new_found = new_summary.get("tags_found", [])

        severities = ("warning", "error")
        if any(
            new_count.get(severity, 0) > reference_count.get(severity, 0)
            for severity in severities
        ):
            status = RegressionAnalysisStatus.REGRESSION
        elif any(
            new_count.get(severity, 0) < reference_count.get(severity, 0)
            for severity in severities
        ):
            status = RegressionAnalysisStatus.IMPROVEMENT
        else:
            status = RegressionAnalysisStatus.STABLE

        details = {
            "added_tags": sorted(set(new_found) - set(reference_found)),
            "removed_tags": sorted(set(reference_found) - set(new_found)),
        }

        return status, details

    def callback_regression_analysis(self) -> bool:
        """
        Analyze regressions compared to reference results.

        This is called once for each architecture, but updates the whole
        analysis for all architectures each time.  This is partly for
        simplicity and robustness (we don't need to work out how to combine
        the new analysis with a previous one), and partly to make it easier
        to handle cases where there isn't a one-to-one mapping between the
        reference results and the new results.
        """
        # Select the newest result for each architecture.
        reference_artifacts = self.find_reference_artifacts("lintian")
        new_artifacts = self.find_new_artifacts(
            "lintian", ArtifactCategory.LINTIAN
        )

        output_data = self.work_request.output_data or OutputData()
        output_data.regression_analysis = {}
        for architecture in sorted(
            set(reference_artifacts) | set(new_artifacts)
        ):
            reference = reference_artifacts.get(architecture)
            reference_source_version, reference_artifact_id, reference_wr = (
                self._extract_artifact_details(reference)
            )
            new = new_artifacts.get(architecture)
            new_source_version, new_artifact_id, new_wr = (
                self._extract_artifact_details(new)
            )

            # A single Lintian task may produce artifacts for multiple
            # architectures (e.g. amd64+all); that may result in regressions
            # or improvements at the level of work request statuses being
            # over-reported against all those architectures rather than just
            # the affected one.  We can live with this since it doesn't
            # change the combined status across all architectures.
            status = self.compare_qa_results(reference_wr, new_wr)
            fine_grained_status, details = self.compare_qa_results_fine_grained(
                reference, new
            )
            if status == RegressionAnalysisStatus.STABLE:
                status = fine_grained_status

            output_data.regression_analysis[architecture] = RegressionAnalysis(
                original_source_version=reference_source_version,
                original_artifact_id=reference_artifact_id,
                new_source_version=new_source_version,
                new_artifact_id=new_artifact_id,
                status=status,
                details=details,
            )

        self.work_request.output_data = output_data
        self.work_request.save()
        return True

    @override
    def get_callback(self, step: str) -> Callable[[], bool]:
        # Current versions of this workflow create a single incremental
        # callback for all architectures called "regression-analysis", but
        # we may be run on workflows that were orchestrated when we created
        # one "regression-analysis-{architecture}" callback per
        # architecture.
        if not step.startswith("regression-analysis"):
            raise NotImplementedError(
                f"Unhandled workflow callback step: {step}"
            )
        return self.callback_regression_analysis



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Workflow to update environments."""

import copy
from dataclasses import dataclass
from typing import Any, cast, override

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import (
    UpdateEnvironmentsWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
    BaseTaskData,
    MmDebstrapData,
    SystemImageBuildData,
)


@dataclass(frozen=True)
class ActionVariables:
    """Variables to pass to an update-collection-with-artifacts action."""

    codename: str
    variant: str | None
    backend: str | None

    def to_dict(self) -> dict[str, str]:
        """Make a dict for an update-collection-with-artifacts action."""
        variables = {"codename": self.codename}
        if self.variant is not None:
            variables["variant"] = self.variant
        if self.backend is not None:
            variables["backend"] = self.backend
        return variables


@dataclass(frozen=True)
class PlannedEnvironment:
    """Relevant data to build environments with a specific task."""

    task_data_template: dict[str, Any]
    architecture: str
    codenames: list[str]
    backends: list[str | None]
    variants: list[str | None]


class UpdateEnvironmentsWorkflow(
    Workflow[UpdateEnvironmentsWorkflowData, BaseDynamicTaskData]
):
    """Build tarballs/images and add them to an environments collection."""

    TASK_NAME = "update_environments"

    def _make_action_variables(
        self,
        *,
        codenames: list[str],
        variants: list[str | None],
        backends: list[str | None],
    ) -> list[ActionVariables]:
        """
        Return collection update action variables for a child work request.

        Each item in the returned list should be used to build an event
        reaction that adds artifacts to a collection.
        """
        action_variables: list[ActionVariables] = []
        variants_or_none: list[str | None] = list(variants)
        if not variants_or_none:
            variants_or_none.append(None)
        backends_or_none: list[str | None] = list(backends)
        if not backends_or_none:
            backends_or_none.append(None)
        for codename in codenames:
            for variant in variants_or_none:
                for backend in backends_or_none:
                    action_variables += [
                        ActionVariables(
                            codename=codename, variant=variant, backend=backend
                        )
                    ]
        return action_variables

    def _add_child(
        self,
        task_name: str,
        task_data_class: type[BaseTaskData],
        template: dict[str, Any],
        *,
        category: ArtifactCategory,
        display_name_prefix: str,
        vendor: str,
        codenames: list[str],
        architecture: str,
        variants: list[str | None],
        backends: list[str | None],
    ) -> None:
        """Create a child work request."""
        # Prepare variables and compute string representations
        codename = codenames[0]
        backend_string = (
            "|".join(b or "no-backend" for b in backends) or "no-backend"
        )
        variant_string = (
            "|".join(v or "no-variant" for v in variants) or "no-variant"
        )
        step = (
            f"{task_name}/{codename}/{architecture}/"
            f"{backend_string}/{variant_string}"
        )
        display_name = (
            f"{display_name_prefix} for {codename} / "
            f"{architecture} / {backend_string} / {variant_string}"
        )

        task_data = copy.deepcopy(template)
        task_data.setdefault("bootstrap_options", {})["architecture"] = (
            architecture
        )
        for repository in task_data.get("bootstrap_repositories", []):
            repository.setdefault("suite", codename)

        # Create the work request with its event reactions
        wr, _ = self.work_request_ensure_child_worker(
            task_name=task_name,
            task_data=task_data_class(**task_data),
            workflow_data=WorkRequestWorkflowData(
                step=step, display_name=display_name
            ),
        )
        for variables in self._make_action_variables(
            codenames=codenames, variants=variants, backends=backends
        ):
            wr.add_event_reaction(
                "on_success",
                ActionUpdateCollectionWithArtifacts(
                    collection=(f"{vendor}@{CollectionCategory.ENVIRONMENTS}"),
                    variables=variables.to_dict(),
                    artifact_filters={"category": category},
                ),
            )

    @override
    def compute_dynamic_data(self) -> BaseDynamicTaskData:
        """
        Compute dynamic data for this workflow.

        :subject: ``vendor``
        :parameter_summary: ``vendor`` being updated and number of targets
        """
        tarballs, images = self._plan_work_requests()
        count = len(tarballs) + len(images)
        summary = f"{self.data.vendor} ({count} environments)"

        return BaseDynamicTaskData(
            subject=self.data.vendor, parameter_summary=summary
        )

    def _plan_work_requests(
        self,
    ) -> tuple[list[PlannedEnvironment], list[PlannedEnvironment]]:
        images = []
        tarballs = []
        for target in self.data.targets:
            codenames = (
                target.codenames
                if isinstance(target.codenames, list)
                else [target.codenames]
            )
            variants: list[str | None] = (
                target.variants
                if isinstance(target.variants, list)
                else [target.variants]
            )
            backends: list[str | None] = (
                cast(list[str | None], target.backends)
                if isinstance(target.backends, list)
                else [target.backends]
            )

            for codename in codenames:
                codename_aliases = [
                    codename,
                    *target.codename_aliases.get(codename, []),
                ]

                for architecture in target.architectures:
                    if target.mmdebstrap_template is not None:
                        environment = PlannedEnvironment(
                            task_data_template=target.mmdebstrap_template,
                            architecture=architecture,
                            codenames=codename_aliases,
                            backends=backends,
                            variants=variants,
                        )
                        tarballs.append(environment)

                    if target.simplesystemimagebuild_template is not None:
                        environment = PlannedEnvironment(
                            task_data_template=target.simplesystemimagebuild_template,
                            architecture=architecture,
                            codenames=codename_aliases,
                            backends=backends,
                            variants=variants,
                        )
                        images.append(environment)

        return tarballs, images

    @override
    def populate(self) -> None:
        """Create child work requests for all targets and architectures."""
        tarballs, images = self._plan_work_requests()

        for tarball in tarballs:
            self._add_child(
                "mmdebstrap",
                MmDebstrapData,
                tarball.task_data_template,
                category=ArtifactCategory.SYSTEM_TARBALL,
                display_name_prefix="Build tarball",
                vendor=self.data.vendor,
                codenames=tarball.codenames,
                architecture=tarball.architecture,
                variants=tarball.variants,
                backends=tarball.backends,
            )

        for image in images:
            self._add_child(
                "simplesystemimagebuild",
                SystemImageBuildData,
                image.task_data_template,
                category=ArtifactCategory.SYSTEM_IMAGE,
                display_name_prefix="Build image",
                vendor=self.data.vendor,
                codenames=image.codenames,
                architecture=image.architecture,
                variants=image.variants,
                backends=image.backends,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Package upload workflow."""

from functools import cached_property
from typing import override

from debusine.artifacts.models import ArtifactCategory
from debusine.client.models import LookupChildType
from debusine.db.models import WorkRequest
from debusine.server.collections.lookup import artifact_ids
from debusine.server.tasks.models import PackageUploadData, PackageUploadInput
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.workflows import (
    Workflow,
    WorkflowValidationError,
    workflow_utils,
)
from debusine.server.workflows.models import (
    PackageUploadWorkflowData,
    PackageUploadWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.signing.tasks.models import DebsignData
from debusine.tasks.models import (
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    LookupSingle,
    MakeSourcePackageUploadData,
    MakeSourcePackageUploadInput,
    MergeUploadsData,
    MergeUploadsInput,
)


class PackageUploadWorkflow(
    Workflow[PackageUploadWorkflowData, PackageUploadWorkflowDynamicData]
):
    """Run package uploads for a set of architectures."""

    TASK_NAME = "package_upload"
    _artifacts_pending_upload: list[tuple[LookupSingle, bool]]

    @override
    def post_init(self) -> None:
        super().post_init()
        self._artifacts_pending_upload: list[tuple[LookupSingle, bool]] = []

    @cached_property
    def source_category(self) -> str | None:
        """Category of source artifact or promise."""
        if self.data.source_artifact is None:
            return None

        return workflow_utils.lookup_result_artifact_category(
            self.work_request.lookup_single(
                self.data.source_artifact,
                expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
            )
        )

    @override
    def validate_input(self) -> None:
        """Raise WorkflowValidationError if needed."""
        if (
            self.data.source_artifact is None
            and self.data.binary_artifacts == LookupMultiple(())
        ):
            raise WorkflowValidationError(
                '"source_artifact" or "binary_artifacts" must be set'
            )

        if self.source_category == ArtifactCategory.SOURCE_PACKAGE and (
            self.data.vendor is None or self.data.codename is None
        ):
            # MakeSourcePackageUpload will be used, but self.data.vendor or
            # self.data.codename are not set.
            raise WorkflowValidationError(
                f'"vendor" and "codename" are required when source '
                f'artifact category is {ArtifactCategory.SOURCE_PACKAGE}'
            )

    @override
    def populate(self) -> None:
        """Create package upload work requests."""
        artifacts_pending_upload: list[tuple[LookupSingle, bool]] = []

        if (source_artifact := self.data.source_artifact) is not None:
            since_version = self.data.since_version
            target_distribution = self.data.target_distribution

            if self.source_category == ArtifactCategory.SOURCE_PACKAGE:
                package_upload_source = (
                    self._populate_make_source_package_upload(
                        source_artifact,
                        since_version=since_version,
                        target_distribution=target_distribution,
                    )
                )
                artifacts_pending_upload.append((package_upload_source, False))

            else:
                artifacts_pending_upload.append((source_artifact, False))

        # Upload binary_artifacts
        results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        artifacts_pending_upload += sorted(
            (
                result.reconstruct(
                    workflow_root=self.work_request.workflow_root
                ),
                True,
            )
            for result in results
        )

        self._artifacts_pending_upload = artifacts_pending_upload

        self._populate_package_uploads()

    @override
    def compute_dynamic_data(self) -> PackageUploadWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: source package names (separated by spaces) of
          ``binary_artifacts`` and ``source_artifact``
        :parameter_summary: package names (same as subject) -> target
        :source_artifact_id: id of source_artifact or None
        :binary_artifacts_ids: ids of the binary_artifacts
        """
        binaries = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        source_package_names = workflow_utils.get_source_package_names(
            binaries,
            configuration_key="binary_artifacts",
            artifact_expected_categories=(
                ArtifactCategory.BINARY_PACKAGE,
                ArtifactCategory.UPLOAD,
            ),
        )

        source_artifact_id = None
        if self.data.source_artifact is not None:
            source_artifact = self.work_request.lookup_single(
                self.data.source_artifact,
                expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
            )

            if source_artifact.artifact is not None:
                source_artifact_id = source_artifact.artifact.id

            source_artifact_names = workflow_utils.get_source_package_names(
                [source_artifact],
                configuration_key="source_artifact",
                artifact_expected_categories=(
                    ArtifactCategory.SOURCE_PACKAGE,
                    ArtifactCategory.UPLOAD,
                ),
            )
            source_package_names = sorted(
                set(source_package_names) | set(source_artifact_names)
            )

        package_names = " ".join(source_package_names)

        return PackageUploadWorkflowDynamicData(
            subject=package_names,
            parameter_summary=f"{package_names} → {self.data.target}",
            source_artifact_id=source_artifact_id,
            binary_artifacts_ids=artifact_ids(binaries),
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        artifacts: list[InputArtifact] = []

        if self.data.source_artifact is not None:
            artifacts.append(
                InputArtifactSingle(
                    lookup=self.data.source_artifact,
                    label="source_artifact",
                    artifact_id=(
                        self.dynamic_data.source_artifact_id
                        if self.dynamic_data is not None
                        else None
                    ),
                )
            )

        artifacts.append(
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            )
        )

        return artifacts

    def _populate_make_source_package_upload(
        self,
        lookup: LookupSingle,
        *,
        since_version: str | None,
        target_distribution: str | None,
    ) -> LookupSingle:
        """
        Create work request for MakeSourcePackageUpload.

        :returns: Lookup of the signed artifact in the collection
        """
        wr, _ = self.work_request_ensure_child_worker(
            task_name="makesourcepackageupload",
            task_data=MakeSourcePackageUploadData(
                input=MakeSourcePackageUploadInput(source_artifact=lookup),
                target_distribution=target_distribution,
                since_version=since_version,
                environment=(
                    f"{self.data.vendor}/match:codename={self.data.codename}"
                ),
                build_architecture=self.data.arch_all_build_architecture,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="Make .changes file",
                step="source-package-upload",
            ),
        )
        self.requires_artifact(wr, lookup)

        artifact_name = "package-upload-source"
        self.provides_artifact(wr, ArtifactCategory.UPLOAD, artifact_name)
        return f"internal@collections/name:{artifact_name}"

    def _populate_signer(
        self,
        signer_identifier: str,
        unsigned_artifact: LookupSingle,
        key: str | None,
        signed_source_artifact: LookupSingle | None,
    ) -> LookupSingle:
        """
        Create a signer: Debsign or ExternalDebsign depending on self.data.

        :param signer_identifier: identifier for the end user of what is going
          to be signed
        :param unsigned_artifact: unsigned artifact for the signer
        :param key: key to be used by the signer. If None, will use
          ExternalDebsign
        :param signed_source_artifact: if not None, signing this artifact
          requires this other signature to have been completed first
        :returns: lookup of the signed artifact
        """
        if key is not None:
            signer, _ = self.work_request_ensure_child_signing(
                task_name="debsign",
                task_data=DebsignData(unsigned=unsigned_artifact, key=key),
                workflow_data=WorkRequestWorkflowData(
                    display_name=f"Sign upload for {signer_identifier}",
                    step=f"debsign-{signer_identifier}",
                ),
            )

        else:
            signer, _ = self.work_request_ensure_child_wait(
                task_name="externaldebsign",
                task_data=ExternalDebsignData(unsigned=unsigned_artifact),
                workflow_data=WorkRequestWorkflowData(
                    display_name=(
                        f"Wait for signature on upload for {signer_identifier}"
                    ),
                    step=f"external-debsign-{signer_identifier}",
                    needs_input=True,
                ),
            )

        self.requires_artifact(signer, unsigned_artifact)
        if signed_source_artifact is not None:
            self.requires_artifact(signer, signed_source_artifact)

        name_signed_artifact = (
            f"package-upload-signed-{unsigned_artifact}".replace(":", "_")
            .replace("@", "_")
            .replace("/", "_")
        )

        self.provides_artifact(
            signer,
            ArtifactCategory.UPLOAD,
            name_signed_artifact,
        )

        return f"internal@collections/name:{name_signed_artifact}"

    def _populate_package_uploads(self) -> None:
        """
        Create work request(s) to upload the artifacts.

        Depending on self.data.merge_uploads will:
        -Create one work request for each self._artifacts_pending_upload
        -Create a single work request of type MergeUpload
        """
        if self.data.merge_uploads:
            wr, _ = self.work_request_ensure_child_worker(
                task_name="mergeuploads",
                task_data=MergeUploadsData(
                    input=MergeUploadsInput(
                        uploads=LookupMultiple.model_validate(
                            [
                                upload
                                for upload, _ in self._artifacts_pending_upload
                            ]
                        )
                    ),
                ),
                workflow_data=WorkRequestWorkflowData(
                    display_name="Merge uploads",
                    step="merge-uploads",
                ),
            )

            any_binaries = False
            for upload, contains_binaries in self._artifacts_pending_upload:
                self.requires_artifact(wr, upload)
                any_binaries = any_binaries or contains_binaries

            merged_artifact_name = (
                f"package-upload-merged-{self.work_request.id}"
            )
            self.provides_artifact(
                wr, ArtifactCategory.UPLOAD, merged_artifact_name
            )

            self._artifacts_pending_upload = [
                (
                    f"internal@collections/name:{merged_artifact_name}",
                    any_binaries,
                )
            ]

        source_upload: LookupSingle | None = None
        source_uploader: WorkRequest | None = None
        for upload, contains_binaries in self._artifacts_pending_upload:
            identifier = str(upload)

            key = self.data.key
            split_binary_signing = False
            if key is None and (
                self.data.binary_key is not None and contains_binaries
            ):
                key = self.data.binary_key
                split_binary_signing = True
            if key is not None or self.data.require_signature:
                upload = self._populate_signer(
                    identifier,
                    upload,
                    key=key,
                    signed_source_artifact=(
                        source_upload if split_binary_signing else None
                    ),
                )
                if not contains_binaries:
                    # There is at most one source artifact to upload.
                    assert source_upload is None
                    source_upload = upload

            uploader, _ = self.work_request_ensure_child_server(
                task_name="packageupload",
                task_data=PackageUploadData(
                    input=PackageUploadInput(upload=upload),
                    target=self.data.target,
                    delayed_days=self.data.delayed_days,
                ),
                workflow_data=WorkRequestWorkflowData(
                    display_name=f"Package upload {identifier}",
                    step=f"package-upload-{identifier}",
                ),
            )
            if not contains_binaries:
                # There is at most one source artifact to upload.
                assert source_uploader is None
                source_uploader = uploader

            self.requires_artifact(uploader, upload)
            if split_binary_signing and source_uploader is not None:
                uploader.add_dependency(source_uploader)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the collections views."""

import datetime as dt
from typing import ClassVar, override
from unittest import mock

from django.db.models import Max
from django.template.response import SimpleTemplateResponse
from django.test import RequestFactory
from django.urls import reverse
from rest_framework import status

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
)
from debusine.db.models import Collection, CollectionItem, Workspace
from debusine.db.models.collections import CollectionRelation
from debusine.db.playground import scenarios
from debusine.test.django import AllowAll, TestCase, override_permission
from debusine.web.views import ui_shortcuts
from debusine.web.views.collections import (
    CollectionCategoryListView,
    CollectionItemDetailView,
    CollectionListView,
)
from debusine.web.views.places import ContainerPlace
from debusine.web.views.tests.utils import ViewTestMixin
from debusine.web.views.ui.collections import CollectionUI


class CollectionViewsTestsBase(ViewTestMixin, TestCase):
    """Base fixture for testing Collection views."""

    scenario = scenarios.DefaultContext(set_current=True)
    default_collections: ClassVar[list[Collection]]
    sid: ClassVar[Collection]
    trixie: ClassVar[Collection]
    environments: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.default_collections = list(cls.scenario.workspace.collections.all())
        cls.sid = cls.playground.create_collection(
            "sid",
            CollectionCategory.SUITE,
        )
        cls.trixie = cls.playground.create_collection(
            "trixie",
            CollectionCategory.SUITE,
        )
        cls.environments = cls.playground.create_collection(
            "debian",
            CollectionCategory.ENVIRONMENTS,
        )


class CollectionListViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionListView`."""

    url: ClassVar[str]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.url = reverse(
            "workspaces:collections:list",
            kwargs={
                "wname": cls.scenario.workspace.name,
            },
        )

    def test_get_place(self) -> None:
        view = CollectionListView()
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertIsInstance(place, ContainerPlace)
        self.assertEqual(place.title, "Collections")
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:list",
                kwargs={"wname": self.scenario.workspace.name},
            ),
        )
        self.assertIsNone(place.icon)
        self.assertEqual(
            place.description, "System public workspace collection list"
        )
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.scenario.workspace.get_absolute_url()
        )

    def test_permissions(self) -> None:
        """Test permissions on the collection list view."""
        self.assertSetsCurrentWorkspace(self.scenario.workspace, self.url)

    def test_list(self) -> None:
        """Test collection list view."""
        response = self.client.get(self.url)
        self.assertResponseHTML(response)
        queryset = response.context["collection_list"]
        self.assertQuerySetEqual(
            queryset,
            [
                *self.default_collections,
                self.sid,
                self.trixie,
                self.environments,
            ],
            ordered=False,
        )

    def test_hides_internal(self) -> None:
        """Internal collections are not shown."""
        internal = self.playground.create_collection(
            "internal",
            CollectionCategory.WORKFLOW_INTERNAL,
            workspace=self.scenario.workspace,
        )
        response = self.client.get(self.url)
        self.assertResponseHTML(response)
        queryset = response.context["collection_list"]
        self.assertFalse(queryset.filter(pk=internal.pk).exists())


class CollectionCategoryListViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionCategoryListView`."""

    def test_get_place(self) -> None:
        view = CollectionCategoryListView()
        view.kwargs = {"ccat": "testcategory"}
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(
            place.title,
            f"{self.scenario.workspace.name} testcategory collections",
        )
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:category_list",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "ccat": "testcategory",
                },
            ),
        )
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_permissions(self) -> None:
        """Test permissions on the collection list view."""
        url = reverse(
            "workspaces:collections:category_list",
            kwargs={
                "wname": self.scenario.workspace.name,
                "ccat": CollectionCategory.SUITE,
            },
        )
        self.assertSetsCurrentWorkspace(self.scenario.workspace, url)

    def test_list_only_selected_category(self) -> None:
        """Test permissions on the collection list view."""
        url = reverse(
            "workspaces:collections:category_list",
            kwargs={
                "wname": self.scenario.workspace.name,
                "ccat": CollectionCategory.SUITE,
            },
        )
        response = self.client.get(url)
        self.assertResponseHTML(response)
        queryset = response.context["collection_list"]
        self.assertQuerySetEqual(
            queryset,
            [self.sid, self.trixie],
            ordered=False,
        )

    def test_hides_internal(self) -> None:
        """Internal collections are not shown."""
        self.playground.create_collection(
            "internal",
            CollectionCategory.WORKFLOW_INTERNAL,
            workspace=self.scenario.workspace,
        )
        url = reverse(
            "workspaces:collections:category_list",
            kwargs={
                "wname": self.scenario.workspace.name,
                "ccat": CollectionCategory.WORKFLOW_INTERNAL,
            },
        )
        response = self.client.get(url)
        self.assertResponseHTML(response)
        queryset = response.context["collection_list"]
        self.assertQuerySetEqual(queryset.all(), [])


class CollectionDetailViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionDetailView`."""

    def test_permissions(self) -> None:
        """Test permissions on the collection list view."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace, self.trixie.get_absolute_url()
        )

    def test_detail(self) -> None:
        """Test collection detail view."""
        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)

        # Check table
        table = self.assertHasElement(
            tree, "//table[@id='collection-details-table']"
        )
        # First row
        element = self.assertHasElement(table, '//tbody/tr[1]/th[1]')
        self.assertTextContentEqual(
            element, "Full history retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[1]/td[1]')
        self.assertTextContentEqual(element, "Always")

        # Second row
        element = self.assertHasElement(table, '//tbody/tr[2]/th[1]')
        self.assertTextContentEqual(
            element, "Metadata only retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[2]/td[1]')
        self.assertTextContentEqual(element, "Always")

        # Third row
        element = self.assertHasElement(table, '//tbody/tr[3]/th[1]')
        self.assertTextContentEqual(element, "Retains artifacts")
        element = self.assertHasElement(table, '//tbody/tr[3]/td[1]')
        self.assertTextContentEqual(element, "Always")

        self.assertFalse(tree.xpath("//div[@id='metadata']"))

    def test_detail_disable_data_retention(self) -> None:
        self.trixie.full_history_retention_period = dt.timedelta(0)
        self.trixie.metadata_only_retention_period = dt.timedelta(0)
        self.trixie.save()

        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(
            tree, "//table[@id='collection-details-table']"
        )

        element = self.assertHasElement(table, '//tbody/tr[1]/th[1]')
        self.assertTextContentEqual(
            element, "Full history retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[1]/td[1]')
        self.assertTextContentEqual(element, "0")

        element = self.assertHasElement(table, '//tbody/tr[2]/th[1]')
        self.assertTextContentEqual(
            element, "Metadata only retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[2]/td[1]')
        self.assertTextContentEqual(element, "0")

    def test_detail_fixed_data_retention(self) -> None:
        self.trixie.full_history_retention_period = dt.timedelta(days=1)
        self.trixie.metadata_only_retention_period = dt.timedelta(days=7)
        self.trixie.save()

        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(
            tree, "//table[@id='collection-details-table']"
        )

        element = self.assertHasElement(table, '//tbody/tr[1]/th[1]')
        self.assertTextContentEqual(
            element, "Full history retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[1]/td[1]')
        self.assertTextContentEqual(element, "1")

        element = self.assertHasElement(table, '//tbody/tr[2]/th[1]')
        self.assertTextContentEqual(
            element, "Metadata only retention period (days)"
        )
        element = self.assertHasElement(table, '//tbody/tr[2]/td[1]')
        self.assertTextContentEqual(element, "7")

    def test_metadata(self) -> None:
        """Test collection metadata view."""
        self.trixie.data = {"test": True}
        self.trixie.save()
        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        metadata = self.assertHasElement(tree, "//div[@id='metadata']")
        self.assertYAMLContentEqual(metadata.div[1], self.trixie.data)

    def test_not_found(self) -> None:
        """Test looking up an invalid collection."""
        response = self.client.get(
            reverse(
                "workspaces:collections:detail",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "ccat": self.trixie.category,
                    "cname": "bookworm",
                },
            )
        )
        self.assertResponseHTML(response, status_code=status.HTTP_404_NOT_FOUND)

    def test_bare_data_stats(self) -> None:
        for i in range(2):
            self.playground.create_bare_data_item(
                self.trixie, f"test{i}", category=BareDataCategory.TEST
            )
        for i in range(3):
            self.playground.create_bare_data_item(
                self.trixie,
                f"log{i}",
                category=BareDataCategory.PACKAGE_BUILD_LOG,
            )
        for i in range(4):
            item = self.playground.create_bare_data_item(
                self.trixie, f"test_deleted_{i}", category=BareDataCategory.TEST
            )
            self.trixie.manager.remove_item(item)
        for i in range(5):
            item = self.playground.create_bare_data_item(
                self.trixie,
                f"log_deleted_{i}",
                category=BareDataCategory.PACKAGE_BUILD_LOG,
            )
            self.trixie.manager.remove_item(item)

        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        stats = self.assertHasElement(
            tree, "//table[@id='stats-bare-data']/tbody"
        )
        contents = [
            (
                self.get_node_text_normalized(tr.th),
                self.get_node_text_normalized(tr.td[0]),
                self.get_node_text_normalized(tr.td[1]),
            )
            for tr in stats.tr
        ]
        self.assertEqual(
            contents,
            [
                ('debian:package-build-log', '3', '5'),
                ('debusine:test', '2', '4'),
            ],
        )
        self.assertEqual(
            stats.tr[0].th.a.get("href"),
            reverse(
                "workspaces:collections:search",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "cname": self.trixie.name,
                    "ccat": self.trixie.category,
                },
            )
            + "?category="
            + BareDataCategory.PACKAGE_BUILD_LOG,
        )
        self.assertEqual(
            stats.tr[1].th.a.get("href"),
            reverse(
                "workspaces:collections:search",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "cname": self.trixie.name,
                    "ccat": self.trixie.category,
                },
            )
            + "?category="
            + BareDataCategory.TEST,
        )

    def test_suite_based_on_relationship(self) -> None:
        suite_based_on_xpath = '//a[@id="suite-based-on"]'

        # The link is not shown when there is no suite-based-on relation.
        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertEqual(tree.xpath(suite_based_on_xpath), [])

        # Create the relationship
        derived_suite = self.playground.create_collection(
            "r-django",
            CollectionCategory.SUITE,
        )

        CollectionRelation.objects.create(
            source=derived_suite,
            target=self.trixie,
            created_by_user=self.scenario.user,
            type=CollectionRelation.Types.SUITE_BASED_ON,
        )

        response = self.client.get(derived_suite.get_absolute_url())
        tree = self.assertResponseHTML(response)

        # Now it's displayed and contains the right link and text
        suite_based_on_a = self.assertHasElement(tree, suite_based_on_xpath)

        self.assertEqual(
            suite_based_on_a.get("href"), self.trixie.get_absolute_url()
        )

        self.assertTextContentEqual(suite_based_on_a, str(self.trixie))

    def test_suite_targeting_relationship(self) -> None:
        suite_targeting_xpath = '//a[@id="suite-targeting"]'

        # The link is not shown when there is no suite-targeting relation.
        response = self.client.get(self.trixie.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertEqual(tree.xpath(suite_targeting_xpath), [])

        # Create the relationship
        suite_targeting = self.playground.create_collection(
            "trixie_backports",
            CollectionCategory.SUITE,
        )

        CollectionRelation.objects.create(
            source=suite_targeting,
            target=self.trixie,
            created_by_user=self.scenario.user,
            type=CollectionRelation.Types.SUITE_TARGETING,
        )

        response = self.client.get(suite_targeting.get_absolute_url())
        tree = self.assertResponseHTML(response)

        # Now it's displayed and contains the right link and text
        suite_targeting_a = self.assertHasElement(tree, suite_targeting_xpath)

        self.assertEqual(
            suite_targeting_a.get("href"), self.trixie.get_absolute_url()
        )

        self.assertTextContentEqual(suite_targeting_a, str(self.trixie))


class CollectionCreateViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionCreateView` class."""

    scenario = scenarios.DefaultContext()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.url = reverse(
            "workspaces:collections:create",
            kwargs={"wname": self.scenario.workspace.name},
        )

    def test_permissions(self) -> None:
        """Test permissions."""
        self.assertSetsCurrentWorkspace(self.scenario.workspace, self.url)
        self.assertEnforcesPermission(
            self.scenario.workspace.can_create_collections,
            self.url,
            "get_context_data",
        )

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_configure(self) -> None:
        response = self.client.post(
            self.url,
            {
                "name": "test",
                "category": CollectionCategory.SUITE,
                "full_history_retention_period": 2,
                "metadata_only_retention_period": 3,
                "data": "created: true",
            },
        )
        coll = Collection.objects.get(
            workspace=self.scenario.workspace,
            name="test",
            category=CollectionCategory.SUITE,
        )
        self.assertRedirects(response, coll.get_absolute_url())

        self.assertEqual(coll.name, "test")
        self.assertEqual(
            coll.retains_artifacts, Collection.RetainsArtifacts.ALWAYS
        )
        self.assertEqual(
            coll.full_history_retention_period, dt.timedelta(days=2)
        )
        self.assertEqual(
            coll.metadata_only_retention_period, dt.timedelta(days=3)
        )
        self.assertEqual(coll.data, {"created": True})

    @override_permission(Workspace, "can_create_collections", AllowAll)
    def test_get(self) -> None:
        response = self.client.get(self.url)
        self.assertResponseHTML(response)
        assert isinstance(response, SimpleTemplateResponse)


class CollectionUpdateViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionUpdateView` class."""

    scenario = scenarios.DefaultContext()

    def url(self, collection: Collection) -> str:
        """Return the update URL for a collection."""
        return reverse(
            "workspaces:collections:update",
            kwargs={
                "wname": collection.workspace.name,
                "cname": collection.name,
                "ccat": collection.category,
            },
        )

    def test_permissions(self) -> None:
        """Test permissions."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace, self.url(self.sid)
        )
        self.assertEnforcesPermission(
            self.sid.can_configure, self.url(self.sid), "get_context_data"
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_configure(self) -> None:
        response = self.client.post(
            self.url(self.sid),
            {
                "name": "test",
                "full_history_retention_period": 2,
                "metadata_only_retention_period": 3,
                "data": "changed: true",
            },
            follow=True,
        )
        self.sid.refresh_from_db()
        self.assertRedirects(response, self.sid.get_absolute_url())

        self.assertEqual(self.sid.name, "test")
        self.assertEqual(
            self.sid.full_history_retention_period, dt.timedelta(days=2)
        )
        self.assertEqual(
            self.sid.metadata_only_retention_period, dt.timedelta(days=3)
        )
        self.assertEqual(self.sid.data, {"changed": True})

        # Configure and delete buttons are displayed
        collection_ui = CollectionUI(RequestFactory().get("/"), self.sid)

        self.assertInHTML(
            collection_ui.place_update.as_button(),
            response.content.decode(),
        )
        self.assertInHTML(
            collection_ui.place_collection_relation_configure.as_button(),
            response.content.decode(),
        )

    @override_permission(Collection, "can_configure", AllowAll)
    def test_initial_form(self) -> None:
        response = self.client.get(self.url(self.sid))
        self.assertResponseHTML(response)
        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        form = response.context_data["form"]
        self.assertIsNone(form["full_history_retention_period"].initial)
        self.assertEqual(form["name"].initial, "sid")


class CollectionDeleteViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionDeleteView` class."""

    scenario = scenarios.DefaultContext()

    def url(self, collection: Collection) -> str:
        """Return the delete URL for a collection."""
        return reverse(
            "workspaces:collections:delete",
            kwargs={
                "wname": collection.workspace.name,
                "cname": collection.name,
                "ccat": collection.category,
            },
        )

    def test_permissions(self) -> None:
        """Test permissions."""
        self.assertSetsCurrentWorkspace(
            self.scenario.workspace, self.url(self.sid)
        )
        self.assertEnforcesPermission(
            self.sid.can_delete, self.url(self.sid), "get_context_data"
        )

    @override_permission(Collection, "can_delete", AllowAll)
    def test_configure(self) -> None:
        with mock.patch(
            "debusine.db.models.Collection.delete_collection", autospec=True
        ) as delete_collection:
            response = self.client.post(self.url(self.sid), {})
        delete_collection.assert_called_once_with(self.sid)
        self.assertRedirects(
            response, self.scenario.workspace.get_absolute_url()
        )

    @override_permission(Collection, "can_delete", AllowAll)
    def test_form_empty_collection(self) -> None:
        response = self.client.get(self.url(self.sid))
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, "//p[@id='collection-is-empty']")

    @override_permission(Collection, "can_delete", AllowAll)
    def test_form_collection_with_items(self) -> None:
        source = self.playground.create_source_artifact()
        self.sid.manager.add_artifact(
            source,
            user=self.scenario.user,
            variables={"component": "main", "section": "extra"},
        )

        response = self.client.get(self.url(self.sid))
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, "//p[@id='collection-contains-items']")


class CollectionSearchViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionSearchView`."""

    items: ClassVar[dict[str, CollectionItem]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        cls.items = {}
        for name in ("foo", "bar"):
            for ver in range(2):
                source = cls.playground.create_source_artifact(
                    name=f"lib{name}{ver}"
                )
                cls.items[f"lib{name}{ver}"] = cls.trixie.manager.add_artifact(
                    source,
                    user=cls.scenario.user,
                    variables={"component": "main", "section": "devel"},
                )
                binary, _ = cls.playground.create_artifact(
                    category=ArtifactCategory.BINARY_PACKAGE,
                    data={
                        "srcpkg_name": source.data["name"],
                        "srcpkg_version": source.data["version"],
                        "deb_fields": {
                            "Package": f"lib{name}{ver}-dev",
                            "Version": source.data["version"],
                            "Architecture": "amd64",
                        },
                        "deb_control_files": [],
                    },
                    paths=[f"lib{name}{ver}-dev_{ver}.0_amd64.deb"],
                    create_files=True,
                    skip_add_files_in_store=True,
                )
                cls.items[f"lib{name}{ver}-dev"] = (
                    cls.trixie.manager.add_artifact(
                        binary,
                        user=cls.scenario.user,
                        variables={
                            "component": "main",
                            "section": "devel",
                            "priority": "optional",
                        },
                    )
                )

        source = cls.playground.create_source_artifact(name="libold0")
        cls.items["libold0"] = cls.trixie.manager.add_artifact(
            source,
            user=cls.scenario.user,
            variables={"component": "main", "section": "devel"},
        )
        cls.trixie.manager.remove_item(
            cls.items["libold0"], user=cls.scenario.user
        )

    def test_permissions(self) -> None:
        """Test permissions on the collection list view."""
        url = reverse(
            "workspaces:collections:search",
            kwargs={
                "wname": self.scenario.workspace.name,
                "ccat": CollectionCategory.SUITE,
                "cname": "trixie",
            },
        )
        self.assertSetsCurrentWorkspace(self.scenario.workspace, url)

    def test_not_found(self) -> None:
        """Test looking up an invalid collection."""
        response = self.client.get(
            reverse(
                "workspaces:collections:search",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "ccat": self.trixie.category,
                    "cname": "bookworm",
                },
            )
        )
        self.assertResponseHTML(response, status_code=status.HTTP_404_NOT_FOUND)

    def test_defaults(self) -> None:
        """Test collection search view."""
        response = self.client.get(self.trixie.get_absolute_url_search())
        tree = self.assertResponseHTML(response)

        table = self.assertHasElement(
            tree, "//table[@id='collection-item-list']"
        )
        self.assertEqual(len(table.tbody.tr), 8)

        ctx = response.context
        request = response.wsgi_request

        items = ctx["paginator"].page_obj.object_list
        self.assertCountEqual(
            items, [v for k, v in self.items.items() if k != "libold0"]
        )

        self.assertEqual(ctx["collection"], self.trixie)

        sample_item = items[0]
        self.assertEqual(
            sample_item.ui(request).ui_shortcuts,
            [
                ui_shortcuts.create_collection_item(sample_item),
                ui_shortcuts.create_artifact_view(sample_item.artifact),
                ui_shortcuts.create_artifact_download(sample_item.artifact),
            ],
        )

        paginator = ctx["paginator"]
        self.assertEqual(paginator.page_obj.number, 1)
        self.assertEqual(paginator.per_page, 50)
        self.assertEqual(paginator.count, 8)
        self.assertEqual(paginator.num_pages, 1)

    def test_ui_shortcuts(self) -> None:
        """Check generated ui shortcuts."""
        response = self.client.get(self.trixie.get_absolute_url_search())
        request = response.wsgi_request
        self.assertResponseHTML(response)
        item = response.context["paginator"].page_obj.object_list[0]
        self.assertEqual(
            item.ui(request).ui_shortcuts,
            [
                ui_shortcuts.create_collection_item(item),
                ui_shortcuts.create_artifact_view(item.artifact),
                ui_shortcuts.create_artifact_download(item.artifact),
            ],
        )

    def test_ui_shortcuts_bare_items(self) -> None:
        """Check generated ui shortcuts for bare items."""
        collection = self.scenario.workspace.get_singleton_collection(
            pc=self.scenario.context_ui(),
            category=CollectionCategory.PACKAGE_BUILD_LOGS,
        )
        data = {
            "work_request_id": 1,
            "vendor": "debian",
            "codename": "sid",
            "architecture": "amd64",
            "srcpkg_name": "hello",
            "srcpkg_version": "1.0-1",
        }
        collection.manager.add_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            user=self.scenario.user,
            data=data,
        )

        response = self.client.get(collection.get_absolute_url_search())
        request = response.wsgi_request
        self.assertResponseHTML(response)
        item = response.context["paginator"].page_obj.object_list[0]
        self.assertEqual(
            item.ui(request).ui_shortcuts,
            [ui_shortcuts.create_collection_item(item)],
        )

    def test_search_name(self) -> None:
        """Test collection search view."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(), data={"name": "libf"}
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertCountEqual(
            items, [v for k, v in self.items.items() if k.startswith("libfoo")]
        )

    def test_search_partial_name_matching(self) -> None:
        """Test collection search name allows partial matching."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(), data={"name": "foo"}
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertCountEqual(
            items, [v for k, v in self.items.items() if "foo" in k]
        )

    def test_search_name_case_sensitive(self) -> None:
        """Test collection search name matching is case-sensitive."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(), data={"name": "Foo"}
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertEqual(list(items), [])

    def test_search_category(self) -> None:
        """Test collection search view."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(),
            data={"category": ArtifactCategory.BINARY_PACKAGE},
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertCountEqual(
            items, [v for k, v in self.items.items() if k.endswith("-dev")]
        )

    def test_search_bad_category(self) -> None:
        """Test collection search with an invalid category."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(),
            data={"category": "does-not-exist"},
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertQuerySetEqual(items, [])

    def test_search_historical(self) -> None:
        """Test collection search view."""
        response = self.client.get(
            self.trixie.get_absolute_url_search(),
            data={"historical": True},
        )
        self.assertResponseHTML(response)
        ctx = response.context
        items = ctx["paginator"].page_obj.object_list
        self.assertCountEqual(items, [self.items["libold0"]])

    def test_search_pagination(self) -> None:
        """Test collection search view pagination."""
        with mock.patch(
            "debusine.web.views.collections"
            ".CollectionSearchView.get_paginate_by",
            return_value=4,
        ):
            response = self.client.get(self.trixie.get_absolute_url_search())
        self.assertResponseHTML(response)

        ctx = response.context

        items = ctx["paginator"].page_obj.object_list
        self.assertEqual(len(items), 4)

        paginator = ctx["paginator"]
        self.assertEqual(paginator.page_obj.number, 1)
        self.assertEqual(paginator.per_page, 4)
        self.assertEqual(paginator.count, 8)
        self.assertEqual(paginator.num_pages, 2)

    def test_search_query_count(self) -> None:
        """Test that collection search makes a reasonable amount of queries."""
        with self.assertNumQueries(14):
            self.client.get(self.trixie.get_absolute_url_search())


class CollectionItemDetailViewTests(CollectionViewsTestsBase):
    """Tests for :py:class:`CollectionItemDetailView`."""

    item: ClassVar[CollectionItem]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Set up a database layout for views."""
        super().setUpTestData()
        artifact = cls.playground.create_source_artifact(name="libfoo0")
        cls.item = cls.trixie.manager.add_artifact(
            artifact,
            user=cls.scenario.user,
            variables={"component": "main", "section": "shlibs"},
        )

    def test_get_place(self) -> None:
        view = CollectionItemDetailView()
        view.object = self.item
        view.request = RequestFactory().get("/")
        place = view.get_place()
        self.assertEqual(place.title, self.item.name)
        self.assertEqual(place.url, self.item.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)

    def test_permissions(self) -> None:
        """Test permissions on the collection list view."""
        url = self.item.get_absolute_url()
        self.assertSetsCurrentWorkspace(self.scenario.workspace, url)

    def test_defaults(self) -> None:
        """Test viewing an item."""
        url = self.item.get_absolute_url()
        response = self.client.get(url)
        tree = self.assertResponseHTML(response)
        metadata = self.assertHasElement(tree, "//div[@id='metadata']")
        self.assertYAMLContentEqual(metadata.div[1], self.item.data)

    def test_no_item_data(self) -> None:
        """Test viewing an item."""
        self.item.data = {}
        self.item.save()
        url = self.item.get_absolute_url()
        response = self.client.get(url)
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//div[@id='metadata']"))

    def test_wrong_item(self) -> None:
        """Test viewing an item."""
        max_id = CollectionItem.objects.aggregate(Max('id'))['id__max'] + 1
        response = self.client.get(
            reverse(
                "workspaces:collections:item_detail",
                kwargs={
                    "wname": self.scenario.workspace.name,
                    "ccat": self.trixie.category,
                    "cname": self.trixie.name,
                    "iid": str(max_id),
                    "iname": "test",
                },
            )
        )
        self.assertResponseHTML(response, status_code=status.HTTP_404_NOT_FOUND)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Lintian work request plugin."""

import copy
import json
import uuid
from pathlib import Path
from typing import ClassVar, override

from django.db.models import F
from django.template.loader import get_template

from debusine.artifacts import LintianArtifact, PackageBuildLog
from debusine.artifacts.models import DebianLintian, DebianLintianSummary
from debusine.db.models import Artifact, FileInArtifact, WorkRequest, Workspace
from debusine.tasks.models import (
    LintianData,
    LintianDynamicData,
    LintianFailOnSeverity,
    LintianInput,
    LookupMultiple,
)
from debusine.test.django import TestCase
from debusine.web.views.lintian import LintianTab, LintianViewWorkRequestPlugin
from debusine.web.views.work_request import Tab


class LintianViewWorkRequestPluginTests(TestCase):
    """Tests for LintianViewWorkRequestPlugin."""

    fail_on_severity: ClassVar[LintianFailOnSeverity]
    package_name: ClassVar[str]
    workspace: ClassVar[Workspace]
    binary_artifact: ClassVar[Artifact]
    source_artifact: ClassVar[Artifact]
    work_request: ClassVar[WorkRequest]
    package_source_file: ClassVar[str]
    package_binary_file_1: ClassVar[str]
    package_binary_file_2: ClassVar[str]

    DESCRIPTION_LINTIAN_RUN_ID = "description-lintian_run"

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.fail_on_severity = LintianFailOnSeverity.WARNING
        cls.workspace = cls.playground.create_workspace(
            name="Public", public=True
        )

        source_artifact, _ = cls.playground.create_artifact(
            paths=["hello.dsc"],
            create_files=True,
            workspace=cls.workspace,
        )
        binary_artifact_1, _ = cls.playground.create_artifact(
            paths=["hello-1.deb"],
            create_files=True,
            workspace=cls.workspace,
        )
        binary_artifact_2, _ = cls.playground.create_artifact(
            paths=["hello-2.deb"],
            create_files=True,
            workspace=cls.workspace,
        )

        cls.work_request = cls.playground.create_worker_task(
            result=WorkRequest.Results.SUCCESS,
            task_data=LintianData(
                input=LintianInput(
                    source_artifact=source_artifact.id,
                    binary_artifacts=LookupMultiple(
                        (binary_artifact_1.id, binary_artifact_2.id)
                    ),
                ),
                fail_on_severity=cls.fail_on_severity,
            ),
            task_name="lintian",
            workspace=cls.workspace,
            use_data=False,
        )
        cls.package_name = "hello"

        cls.package_source_file = "hello_2.10-3.dsc"
        cls.package_binary_file_1 = "hello_2.10-3.deb"
        cls.package_binary_file_2 = "libhello_2.10-3.deb"

        summary_binary_package = DebianLintian(
            architecture="all",
            summary=DebianLintianSummary(
                package_filename={
                    "hello": cls.package_binary_file_1,
                    "libhello": cls.package_binary_file_2,
                },
                tags_found=[],
                overridden_tags_found=[],
                lintian_version="2.116.3",
                distribution="debian/bookworm",
                tags_count_by_severity={},
            ),
        ).model_dump()

        cls.binary_artifact, _ = cls.playground.create_artifact(
            paths=["summary.json", "lintian.txt"],
            create_files=True,
            workspace=cls.workspace,
        )
        cls.binary_artifact.category = LintianArtifact._category
        cls.binary_artifact.data = summary_binary_package
        cls.binary_artifact.created_by_work_request = cls.work_request
        cls.binary_artifact.save()

        summary_source_package = DebianLintian(
            architecture="source",
            summary=DebianLintianSummary(
                package_filename={
                    "hello": cls.package_source_file,
                },
                tags_found=[],
                overridden_tags_found=[],
                lintian_version="2.116.3",
                distribution="debian/bookworm",
                tags_count_by_severity={},
            ),
        ).model_dump()

        cls.source_artifact, _ = cls.playground.create_artifact(
            paths=["lintian.txt"],
            create_files=True,
            workspace=cls.workspace,
        )
        cls.source_artifact.category = LintianArtifact._category
        cls.source_artifact.data = summary_source_package
        cls.source_artifact.created_by_work_request = cls.work_request
        cls.source_artifact.save()

        build_log_artifact, _ = cls.playground.create_artifact(
            paths=["some-file.log"],
            create_files=True,
            workspace=cls.workspace,
        )
        build_log_artifact.category = PackageBuildLog._category
        build_log_artifact.save()

    def add_file_in_artifact(
        self,
        artifact: Artifact,
        path: str,
        contents: bytes,
        complete: bool = True,
    ) -> None:
        """
        Add file in the artifact.

        :param artifact: artifact that the file is added to
        :param path: path in the artifact
        :param contents: contents written into the file that is added in the
          artifact
        """
        file_backend = artifact.workspace.scope.file_stores.order_by(
            F("filestoreinscope__upload_priority").desc(nulls_last=True)
        )[0].get_backend_object()
        if complete:
            fileobj = self.playground.create_file_in_backend(
                file_backend, contents
            )
        else:
            fileobj = self.playground.create_file(contents)
        FileInArtifact.objects.create(
            artifact=artifact, path=path, file=fileobj, complete=complete
        )

    def create_analysis(self) -> Path:
        """Create an analysis file with tags (without the summary)."""
        contents = json.dumps(
            {
                "summary": {"package_filename": {"hello": "hello_2.10-3.dsc"}},
                "tags": [
                    {
                        "severity": "pedantic",
                        "package": "hello",
                        "tag": "license-problem-gfdl-non-official-text",
                        "explanation": "The given source file is licensed...",
                        "comment": "this is a comment",
                        "note": "invariant part is\\n: with no invariant...",
                        "pointer": "debian/copyright",
                    },
                    {
                        "severity": "pedantic",
                        "package": "hello",
                        "tag": "license-problem-gfdl-non-official-text",
                        "explanation": "The given source file is licensed...",
                        "comment": "",
                        "note": "invariant part is: with no invariant...",
                        "pointer": "doc/hello.info",
                    },
                    {
                        "severity": "warning",
                        "package": "hello",
                        "tag": "some-other-tag",
                        "explanation": "",
                        "comment": "",
                        "note": "some-other-tag note...",
                        "pointer": "src/some-other-tag-file",
                    },
                ],
            }
        ).encode("utf-8")
        return self.create_temporary_file(contents=contents)

    def get_tab(
        self, source: WorkRequest | LintianViewWorkRequestPlugin
    ) -> LintianTab:
        """Return the tab for a work request or a plugin."""
        if isinstance(source, WorkRequest):
            source = LintianViewWorkRequestPlugin(source.get_task())
        tabs: list[Tab] = []
        source.add_tabs(tabs)
        self.assertEqual(len(tabs), 1)
        tab = tabs[0]
        assert isinstance(tab, LintianTab)
        self.assertEqual(tab.name, "lintian")
        self.assertEqual(tab.label, "Lintian")
        return tab

    def test_get_context_data(self) -> None:
        """Test get_context_data() return the expected context data."""
        lintian_txt_path = (
            self.binary_artifact.get_absolute_url_download() + "lintian.txt"
        )
        plugin = LintianViewWorkRequestPlugin(self.work_request.get_task())
        context_data = self.get_tab(plugin).get_context_data()

        # Verify some keys in separate methods
        context_data.pop("tags")
        context_data.pop("tags_count_severity")
        context_data.pop("request_data")

        self.assertEqual(
            context_data,
            {
                "result": self.work_request.result,
                "lintian_txt_path": lintian_txt_path,
                "fail_on_severity": self.fail_on_severity,
            },
        )

        self.assertEqual(
            plugin.get_context_data(),
            {
                "description_data": {},
                "description_template": "web/_lintian-description.html",
            },
        )

    def test_get_context_data_verify_tag_count_severity(self) -> None:
        """Test get_context_data(): verify "tags_count_severity"."""
        self.binary_artifact.data["summary"] = {
            "package_filename": {"hello": "hello_2.10-3.deb"},
            "tags_count_by_severity": {
                "classification": 0,
                "error": 0,
                "experimental": 0,
                "info": 0,
                "overridden": 0,
                "pedantic": 2,
                "warning": 1,
            },
        }
        self.binary_artifact.save()

        self.source_artifact.data["summary"] = {
            "package_filename": {"hello": "hello_2.10-3.dsc"},
            "tags_count_by_severity": {
                "classification": 0,
                "error": 0,
                "experimental": 0,
                "info": 0,
                "overridden": 5,
                "pedantic": 1,
                "warning": 0,
            },
        }
        self.source_artifact.save()

        context_data = self.get_tab(self.work_request).get_context_data()
        self.assertEqual(
            context_data["tags_count_severity"],
            {
                "classification": 0,
                "error": 0,
                "experimental": 0,
                "info": 0,
                "overridden": 5,
                "pedantic": 3,
                "warning": 1,
            },
        )

    def test_get_context_data_verify_tags(self) -> None:
        """Test get_context_data(): verify "tags"."""
        self.add_file_in_artifact(
            self.binary_artifact,
            "analysis.json",
            self.create_analysis().read_bytes(),
        )

        context_data = self.get_tab(self.work_request).get_context_data()

        expected = {
            "hello_2.10-3.dsc": {
                "license-problem-gfdl-non-official-text": {
                    "explanation": "The given source file is licensed...",
                    "severity": "pedantic",
                    "occurrences": [
                        {
                            "pointer": "debian/copyright",
                            "note": "invariant part is\n: with no invariant...",
                            "comment": "this is a comment",
                        },
                        {
                            "pointer": "doc/hello.info",
                            "note": "invariant part is: with no invariant...",
                            "comment": "",
                        },
                    ],
                },
                "some-other-tag": {
                    "explanation": "",
                    "severity": "warning",
                    "occurrences": [
                        {
                            "pointer": "src/some-other-tag-file",
                            "note": "some-other-tag note...",
                            "comment": "",
                        }
                    ],
                },
            }
        }

        actual = {
            filename: {
                tag_name: tag.model_dump() for tag_name, tag in tags.items()
            }
            for filename, tags in context_data["tags"].items()
        }

        uuid1 = actual["hello_2.10-3.dsc"][
            "license-problem-gfdl-non-official-text"
        ].pop("uuid")
        self.assertIsInstance(uuid1, uuid.UUID)

        uuid2 = actual["hello_2.10-3.dsc"]["some-other-tag"].pop("uuid")
        self.assertIsInstance(uuid2, uuid.UUID)

        self.assertEqual(actual, expected)

    def test_get_context_data_verify_request(self) -> None:
        """Test get_context_data(): verify "request_data"."""
        task_data_input = self.work_request.task_data["input"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_source_artifact_id=task_data_input["source_artifact"],
            input_binary_artifacts_ids=task_data_input["binary_artifacts"],
        ).model_dump(exclude_unset=True)
        self.work_request.save()

        context_data = self.get_tab(self.work_request).get_context_data()

        expected = {
            "source_artifact": {
                "files": ["hello.dsc"],
                "id": task_data_input["source_artifact"],
                "lookup": task_data_input["source_artifact"],
                "artifact": Artifact.objects.get(
                    id=task_data_input["source_artifact"]
                ),
            },
            "binary_artifacts": {
                "lookup": task_data_input["binary_artifacts"],
                "artifacts": [
                    {
                        "files": ["hello-1.deb"],
                        "id": task_data_input["binary_artifacts"][0],
                        "artifact": Artifact.objects.get(
                            id=task_data_input["binary_artifacts"][0]
                        ),
                    },
                    {
                        "files": ["hello-2.deb"],
                        "id": task_data_input["binary_artifacts"][1],
                        "artifact": Artifact.objects.get(
                            id=task_data_input["binary_artifacts"][1]
                        ),
                    },
                ],
            },
        }

        self.assertEqual(context_data["request_data"], expected)

    def test_get_context_data_no_dynamic_task_data(self) -> None:
        """If there is no dynamic data, context data omits artifact info."""
        task_data_input = self.work_request.task_data["input"]

        context_data = self.get_tab(self.work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {
                "lookup": task_data_input["source_artifact"],
                "id": None,
                "files": [],
            },
        )
        self.assertEqual(
            context_data["request_data"]["binary_artifacts"],
            {"lookup": task_data_input["binary_artifacts"], "artifacts": []},
        )

    def test_get_context_data_missing_source_artifact(self) -> None:
        """If the source artifact is missing, context data omits its files."""
        task_data_input = self.work_request.task_data["input"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_source_artifact_id=task_data_input["source_artifact"],
            input_binary_artifacts_ids=task_data_input["binary_artifacts"],
        ).model_dump(exclude_unset=True)
        self.work_request.save()
        source_artifact_id = task_data_input["source_artifact"]
        source_artifact = Artifact.objects.get(id=source_artifact_id)
        source_artifact.files.clear()
        source_artifact.delete()
        context_data = self.get_tab(self.work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {"lookup": source_artifact_id, "id": None, "files": []},
        )

    def test_get_context_data_missing_binary_artifact(self) -> None:
        """If a binary artifact is missing, context data omits its files."""
        task_data_input = self.work_request.task_data["input"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_source_artifact_id=task_data_input["source_artifact"],
            input_binary_artifacts_ids=task_data_input["binary_artifacts"],
        ).model_dump(exclude_unset=True)
        self.work_request.save()
        binary_artifact_0_id = task_data_input["binary_artifacts"][0]
        binary_artifact_0 = Artifact.objects.get(id=binary_artifact_0_id)
        binary_artifact_0.files.clear()
        binary_artifact_0.delete()
        context_data = self.get_tab(self.work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["binary_artifacts"],
            {
                "lookup": task_data_input["binary_artifacts"],
                "artifacts": [
                    {
                        "files": [],
                        "id": task_data_input["binary_artifacts"][0],
                        "artifact": None,
                    },
                    {
                        "files": ["hello-2.deb"],
                        "id": task_data_input["binary_artifacts"][1],
                        "artifact": Artifact.objects.get(
                            id=task_data_input["binary_artifacts"][1]
                        ),
                    },
                ],
            },
        )

    def test_get_context_data_skips_incomplete_files(self) -> None:
        """get_context_data() skips incomplete files."""
        self.source_artifact.fileinartifact_set.update(complete=False)
        self.binary_artifact.fileinartifact_set.update(complete=False)
        self.add_file_in_artifact(
            self.binary_artifact,
            "analysis.json",
            self.create_analysis().read_bytes(),
            complete=False,
        )

        context_data = self.get_tab(self.work_request).get_context_data()
        self.assertIsNone(context_data["lintian_txt_path"])
        self.assertEqual(context_data["tags"], {})

    def test_template_output(self) -> None:
        """Generic output of the template."""
        task_data_input = self.work_request.task_data["input"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_source_artifact_id=task_data_input["source_artifact"],
            input_binary_artifacts_ids=task_data_input["binary_artifacts"],
        ).model_dump(exclude_unset=True)
        self.work_request.save()
        self.add_file_in_artifact(
            self.binary_artifact,
            "analysis.json",
            self.create_analysis().read_bytes(),
        )

        response = self.client.get(self.work_request.get_absolute_url())

        result_output = get_template("web/_work_request-result.html").render(
            {"result": self.work_request.result}
        )
        self.assertContains(response, f"Result: {result_output}", html=True)
        self.assertContains(
            response, f"Fail on: {self.fail_on_severity}", html=True
        )

        lintian_txt_path = (
            self.binary_artifact.get_absolute_url_download() + "lintian.txt"
        )
        self.assertContains(
            response,
            f'<a href="{lintian_txt_path}">Lintian output</a>',
            html=True,
        )

        # Test code contains the list (testing part of the list) of
        # the summary of the tags
        self.assertContains(
            response,
            "<li>error: 0</li>",
            html=True,
        )
        self.assertContains(
            response,
            "<li>warning: 0</li>",
            html=True,
        )

        task_data_input = self.work_request.task_data["input"]

        tag = response.context["tags"]["hello_2.10-3.dsc"][
            "license-problem-gfdl-non-official-text"
        ].occurrences[0]

        tag_output = get_template("web/_lintian_tag.html").render({"tag": tag})
        self.assertContains(response, f"<li>{tag_output}</li>", html=True)

    def test_template_output_include_note_pointer_comment(self) -> None:
        """Template output of _lintian_tag.html include all the information."""
        tag = {
            "note": "invariant part is:\nwith no invariant",
            "pointer": "debian/copyright",
            "comment": "upstream",
        }

        rendered = get_template("web/_lintian_tag.html").render({"tag": tag})

        self.assertIn("invariant part is:<br>with no invariant", rendered)
        self.assertIn(f"<code>[{tag['pointer']}]</code>", rendered)
        self.assertIn("Comment:", rendered)
        self.assertIn(f"<pre>{tag['comment']}</pre>", rendered)

    def test_template_output_do_not_include_non_needed_information(
        self,
    ) -> None:
        """Template _lintian_tag.html does not output note, comment, pointer."""
        tag = {"note": "", "pointer": "", "comment": ""}
        rendered = get_template("web/_lintian_tag.html").render({"tag": tag})

        self.assertIn("Lintian did not output a note for this tag", rendered)
        self.assertNotIn("<code>", rendered)
        self.assertNotIn("Comment:", rendered)
        self.assertNotIn("<pre></pre>", rendered)

    def test_template_output_no_source_artifact(self) -> None:
        """The view renders OK if there was no source artifact in the input."""
        del self.work_request.task_data["input"]["source_artifact"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_binary_artifacts_ids=self.work_request.task_data["input"][
                "binary_artifacts"
            ],
        ).model_dump(exclude_unset=True)
        self.work_request.save()

        response = self.client.get(self.work_request.get_absolute_url())

        self.assertNotContains(response, "Source artifact", html=True)

    def test_template_output_no_binary_artifacts(self) -> None:
        """The view renders OK if there was no binary artifact in the input."""
        del self.work_request.task_data["input"]["binary_artifacts"]
        self.work_request.dynamic_task_data = LintianDynamicData().model_dump(
            exclude_unset=True
        )
        self.work_request.save()

        response = self.client.get(self.work_request.get_absolute_url())
        self.assertNotContains(response, "Binary artifact", html=True)

    def test_find_lintian_txt_url_path_no_artifacts(self) -> None:
        """_find_lintian_txt_url_path() with no artifacts: return None."""
        self.assertIsNone(LintianTab._find_lintian_txt_url_path([]))

    def test_find_lintian_txt_url_path_artifact_no_files(self) -> None:
        """_find_lintian_txt_url_path(), artifact no files: return None."""
        artifact, _ = self.playground.create_artifact()

        self.assertIsNone(LintianTab._find_lintian_txt_url_path([artifact]))

    def test_template_output_no_lintian_txt(self) -> None:
        """The view renders OK if no lintian.txt was output."""
        self.source_artifact.fileinartifact_set.all().delete()
        self.binary_artifact.fileinartifact_set.all().delete()

        response = self.client.get(self.work_request.get_absolute_url())
        self.assertNotContains(response, "Lintian output", html=True)

    def test_show_configured_task_data(self) -> None:
        """The view shows configured task data if present."""
        orig_source_id = self.work_request.task_data["input"]["source_artifact"]
        configured_source_id = orig_source_id + 42

        # Try unconfigured
        self.work_request.configured_task_data = None
        self.work_request.dynamic_task_data = None
        self.work_request.save()

        response = self.client.get(self.work_request.get_absolute_url())
        self.assertEqual(
            response.context["request_data"]["source_artifact"]["lookup"],
            orig_source_id,
        )

        # Try configured
        self.work_request.configured_task_data = copy.deepcopy(
            self.work_request.task_data
        )
        self.work_request.configured_task_data["input"]["source_artifact"] = (
            configured_source_id
        )
        self.work_request.save()
        response = self.client.get(self.work_request.get_absolute_url())
        self.assertEqual(
            response.context["request_data"]["source_artifact"]["lookup"],
            configured_source_id,
        )

    def test_description_data(self) -> None:
        self.work_request.task_data["include_tags"] = [
            "included-1",
            "included-2",
        ]
        self.work_request.task_data["exclude_tags"] = [
            "excluded-1",
            "excluded-2",
        ]

        # Set some dynamic_data
        task_data_input = self.work_request.task_data["input"]
        self.work_request.dynamic_task_data = LintianDynamicData(
            input_source_artifact_id=task_data_input["source_artifact"],
            input_binary_artifacts_ids=task_data_input["binary_artifacts"],
            environment_id=0,
            subject="hello",
        ).model_dump(exclude_unset=True)
        self.work_request.save()

        response = self.client.get(self.work_request.get_absolute_url())

        # Check contents of context
        ctx = response.context
        description_data, description_template = (
            ctx["description_data"],
            ctx["description_template"],
        )
        self.assertEqual(
            description_data,
            {
                "source_artifact_id": task_data_input["source_artifact"],
                "binary_artifacts_ids": task_data_input["binary_artifacts"],
                "environment_id": 0,
                "exclude_tags": ["excluded-1", "excluded-2"],
                "include_tags": ["included-1", "included-2"],
                "fail_on_severity": "warning",
                "build_architecture": None,
                "package_name": "hello",
                "target_distribution": "debian:unstable",
            },
        )
        self.assertEqual(description_template, "web/_lintian-description.html")

        # Check HTML output
        tree = self.assertResponseHTML(response)

        lintian_run = self.assertHasElement(
            tree, f"//p[@id='{self.DESCRIPTION_LINTIAN_RUN_ID}']"
        )
        self.assertTextContentEqual(
            lintian_run, "Lintian run for hello package."
        )

        source_packages = self.assertHasElement(
            tree, "//li[@id='description-source_artifacts']"
        )
        self.assertTextContentEqual(
            source_packages, "Source packages: debusine:test"
        )

        binary_packages = self.assertHasElement(
            tree, "//li[@id='description-binary_artifacts']"
        )
        self.assertTextContentEqual(
            binary_packages, "Binary packages: debusine:test, debusine:test"
        )

        environment = self.assertHasElement(
            tree, "//li[@id='description-environment']"
        )
        self.assertTextContentEqual(
            environment, "Executed in environment 0 (deleted)"
        )

        fail_on_severity = self.assertHasElement(
            tree, "//li[@id='description-fail_on_severity']"
        )
        self.assertTextContentEqual(
            fail_on_severity, "Fail on severity: warning"
        )

        build_architecture = self.assertHasElement(
            tree, "//li[@id='description-build_architecture']"
        )
        self.assertTextContentEqual(
            build_architecture, "Build architecture: auto"
        )

        target_distribution = self.assertHasElement(
            tree, "//li[@id='description-target_distribution']"
        )
        self.assertTextContentEqual(
            target_distribution, "Target distribution: debian:unstable"
        )

        include_tags = self.assertHasElement(
            tree, "//li[@id='description-include_tags']"
        )
        self.assertTextContentEqual(
            include_tags, "Include tags: included-1 included-2"
        )

        exclude_tags = self.assertHasElement(
            tree, "//li[@id='description-exclude_tags']"
        )
        self.assertTextContentEqual(
            exclude_tags, "Exclude tags: excluded-1 excluded-2"
        )

    def test_description_data_not_configured(self) -> None:
        response = self.client.get(self.work_request.get_absolute_url())

        # Check contents of context
        ctx = response.context
        description_data, description_template = (
            ctx["description_data"],
            ctx["description_template"],
        )

        self.assertEqual(description_data, {})
        self.assertEqual(description_template, "web/_lintian-description.html")

        # Does not contain description related information because of no data
        tree = self.assertResponseHTML(response)
        self.assertFalse(
            tree.xpath(f"//p[@id='{self.DESCRIPTION_LINTIAN_RUN_ID}']")
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for workflow Celery integration."""

import datetime as dt
import logging
from itertools import chain, repeat
from typing import Any
from unittest import mock

from django.db import connections, transaction
from django.utils import timezone
from django_prometheus.testutils import assert_metric_diff, save_registry

from debusine.db.models import WorkRequest
from debusine.server.workflows.celery import run_workflow_task
from debusine.tasks.models import OutputData, OutputDataError
from debusine.test.django import TestCase, TransactionTestCase


class RunWorkflowTaskTests(TestCase):
    """Unit tests for :py:func:`run_workflow_task`."""

    def test_no_work_request(self) -> None:
        """The Celery task fails if the work request does not exist."""
        work_request_id = self.playground.create_worker_task().id + 1

        with self.assertLogsContains(
            f"Work request {work_request_id} does not exist",
            logger="debusine.server.workflows.celery",
            level=logging.ERROR,
        ):
            result = run_workflow_task.apply(args=(work_request_id,))

        with self.assertRaises(WorkRequest.DoesNotExist):
            result.get()

    def test_not_running(self) -> None:
        """The Celery task returns False if the work request is not running."""
        wr = self.playground.create_workflow(task_name="noop")

        with self.assertLogsContains(
            f"Error running work request Workflow/noop ({wr.id}): "
            f"Work request is in status pending, not running",
            logger="debusine.server.workflows.celery",
            level=logging.WARNING,
        ):
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        wr.refresh_from_db()
        self.assertIsNotNone(wr.completed_at)
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=(
                            "Work request is in status pending, not running"
                        ),
                        code="wrong-status",
                    )
                ]
            ),
        )

    def test_workflow_run_error(self) -> None:
        """The Celery task returns False if the orchestrator fails."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        with (
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"Orchestrator failed: Boom",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate",
                side_effect=Exception("Boom"),
            ),
        ):
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        wr.refresh_from_db()
        self.assertIsNotNone(wr.completed_at)
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=("Orchestrator failed: Boom"),
                        code="orchestrator-failed",
                    )
                ]
            ),
        )

    def test_sub_workflow(self) -> None:
        """The Celery task returns False if given a sub-workflow."""
        root_wr = self.playground.create_workflow(task_name="noop")
        sub_wr = self.playground.create_workflow(
            task_name="noop", parent=root_wr, mark_running=True
        )

        with self.assertLogsContains(
            f"Error running work request Workflow/noop ({sub_wr.id}): Must be "
            f"populated by its parent workflow instead",
            logger="debusine.server.workflows.celery",
            level=logging.ERROR,
        ):
            result = run_workflow_task.apply(args=(sub_wr.id,))

        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        sub_wr.refresh_from_db()
        self.assertIsNotNone(sub_wr.started_at)
        self.assertIsNotNone(sub_wr.completed_at)
        self.assertEqual(sub_wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(sub_wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            sub_wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=(
                            "Must be populated by its parent workflow instead"
                        ),
                        code="non-root-workflow",
                    )
                ]
            ),
        )

    def test_workflow_callback(self) -> None:
        """A workflow callback is run and marked as completed."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        self.playground.advance_work_request(wr, mark_running=True)

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback"
        ) as mock_noop_callback:
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())
        mock_noop_callback.assert_called_once_with(wr)
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow(self) -> None:
        """A workflow is populated and left running."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def populate() -> None:
            wr.create_child_worker("noop")

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=populate,
        ) as mock_noop_populate:
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())
        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)

    def test_workflow_empty(self) -> None:
        """An empty workflow is populated and marked as completed."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate"
        ) as mock_noop_populate:
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())
        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_emits_metric(self) -> None:
        registry = save_registry()
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        now = timezone.now()

        with (
            mock.patch(
                "django.utils.timezone.now",
                side_effect=chain([now], repeat(now + dt.timedelta(minutes=4))),
            ),
            mock.patch("debusine.server.workflows.noop.NoopWorkflow.populate"),
        ):
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())
        labels = {"task_name": wr.task_name}
        assert_metric_diff(
            registry,
            0,
            "debusine_workflow_orchestrator_duration_bucket",
            **labels,
            le="60.0",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_workflow_orchestrator_duration_bucket",
            **labels,
            le="300.0",
        )
        assert_metric_diff(
            registry,
            1,
            "debusine_workflow_orchestrator_duration_bucket",
            **labels,
            le="+Inf",
        )


class RunWorkflowTaskTransactionTests(TransactionTestCase):
    """Test transactional behaviour of :py;func:`run_workflow_task`."""

    def test_runs_in_transaction(self) -> None:
        """The workflow orchestrator is run in a transaction."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def fake_populate() -> None:
            self.assertTrue(connections["default"].in_atomic_block)

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=fake_populate,
        ) as mock_noop_populate:
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())
        mock_noop_populate.assert_called_once_with()

    def test_locks_workflow(self) -> None:
        with transaction.atomic():
            wr = self.playground.create_workflow()
            wr.mark_running()
            wr_child = wr.create_child_worker(task_name="noop")
            unrelated = self.playground.create_workflow()

        def mock_orchestrate_workflow(*args: Any, **kwargs: Any) -> bool:
            self.assert_work_requests_locked(
                locked=[wr, wr_child], unlocked=[unrelated]
            )
            return True

        with (
            transaction.atomic(),
            mock.patch(
                "debusine.server.workflows.celery.orchestrate_workflow",
                side_effect=mock_orchestrate_workflow,
            ),
        ):
            result = run_workflow_task.apply(args=(wr.id,))

        self.assertTrue(result.get())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for urls module."""

import re
from typing import Any

import pytest

from debusine.test import TestCase
from debusine.utils.urls import URLPath, parse_timeout


class URLPathTests(TestCase):
    """Tests for the :py:class:`URLPath` class."""

    def test_init(self) -> None:
        for args, exp_value, exp_is_file, exp_str in (
            ([""], [], False, "/"),
            (["/"], [], False, "/"),
            (["//"], [], False, "/"),
            (["foo"], ["foo"], True, "foo"),
            (["foo/"], ["foo"], False, "foo/"),
            (["foo/bar.html"], ["foo", "bar.html"], True, "foo/bar.html"),
            (["foo/bar.html/"], ["foo", "bar.html"], False, "foo/bar.html/"),
            (["/foo//bar/"], ["foo", "bar"], False, "foo/bar/"),
            (["/foo//bar/", True], ["foo", "bar"], True, "foo/bar"),
            (["/foo//bar", False], ["foo", "bar"], False, "foo/bar/"),
        ):
            with self.subTest(args=args):
                upath = URLPath(*args)
                self.assertEqual(upath.path, exp_value)
                self.assertEqual(upath.is_file, exp_is_file)
                self.assertEqual(str(upath), exp_str)

    def test_sequence(self) -> None:
        upath = URLPath("/foo/bar.html")

        # Container
        self.assertIn("foo", upath)
        self.assertNotIn("bar", upath)

        # Sized
        self.assertEqual(len(upath), 2)

        # Iterable
        self.assertEqual(list(upath), ["foo", "bar.html"])

        # Sequence
        self.assertEqual(upath[0], "foo")
        self.assertEqual(upath[1], "bar.html")
        self.assertEqual(upath[0:2], ["foo", "bar.html"])

    def test_add(self) -> None:
        for first, second, expected in (
            ("", "", "/"),
            ("/foo.html", "bar.html", "bar.html"),
            ("/foo.html", "api/", "api/"),
            ("/api/", "/send", "api/send"),
            ("/api/", "/send/", "api/send/"),
        ):
            with self.subTest(first=str(first), second=str(second)):
                # Add URLPath
                self.assertEqual(
                    str(URLPath(first) + URLPath(second)), expected
                )

                # In-place add URLPath
                p = URLPath(first)
                p += URLPath(second)
                self.assertEqual(str(p), expected)

                # Add str
                self.assertEqual(str(URLPath(first) + second), expected)

                # In-place add str
                p = URLPath(first)
                p += second
                self.assertEqual(str(p), expected)

    def test_add_invalid_type(self) -> None:
        upath = URLPath()
        value: Any
        for value in (None, 42, (), [], {}):
            with self.subTest(value=repr(value)):
                with self.assertRaisesRegex(
                    TypeError, re.escape("unsupported operand type(s) for +")
                ):
                    upath + value
                with self.assertRaisesRegex(
                    TypeError, re.escape("unsupported operand type(s) for +=")
                ):
                    upath += value

    def test_bool(self) -> None:
        for value, expected in (
            ("", False),
            ("/", False),
            ("/api", True),
            ("/api/", True),
        ):
            with self.subTest(value=value):
                self.assertEqual(bool(URLPath(value)), expected)

    def test_pop(self) -> None:
        for value, expected in (
            ("", "/"),
            ("/", "/"),
            ("/api", "/"),
            ("/api/", "/"),
            ("/foo/bar.html", "foo/"),
            ("/foo/bar/", "foo/"),
        ):
            with self.subTest(value=value):
                upath = URLPath(value)
                upath.pop()
                self.assertEqual(str(upath), expected)


def test_parse_timeout() -> None:
    """Test parse_timeout utility."""
    assert parse_timeout("30") == 30
    assert parse_timeout("10, 20") == (10, 20)
    assert parse_timeout(" 5 , 15 ") == (5, 15)

    with pytest.raises(ValueError, match="Invalid timeout value"):
        parse_timeout("invalid")
    with pytest.raises(ValueError, match="Invalid timeout value"):
        parse_timeout("10, invalid")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for artifacts templatetags."""

from debusine.db.models import Artifact
from debusine.db.models.artifacts import (
    ARTIFACT_CATEGORY_ICON_NAMES,
    ARTIFACT_CATEGORY_SHORT_NAMES,
)
from debusine.test.django import TestCase
from debusine.web.templatetags.artifacts import (
    artifact_category_label,
    artifact_icon_name,
    artifact_link,
    artifact_links,
)


class ArtifactTagsTests(TestCase):
    """Tests for artifacts tags."""

    def test_category_label_str(self) -> None:
        """Test artifact_category_label tag on strings."""
        for k, v in ARTIFACT_CATEGORY_SHORT_NAMES.items():
            self.assertEqual(artifact_category_label(k), v)
        self.assertEqual(artifact_category_label("debusine:test"), "artifact")

    def test_category_label_artifact(self) -> None:
        """Test artifact_category_label tag on artifacts."""
        for k, v in ARTIFACT_CATEGORY_SHORT_NAMES.items():
            artifact = Artifact(category=k)
            self.assertEqual(artifact_category_label(artifact), v)
        artifact = Artifact(category="debusine:test")
        self.assertEqual(artifact_category_label(artifact), "artifact")

    def test_icon_name_str(self) -> None:
        """Test artifact_icon_name tag on strings."""
        for k, v in ARTIFACT_CATEGORY_ICON_NAMES.items():
            self.assertEqual(artifact_icon_name(k), v)
        self.assertEqual(artifact_icon_name("debusine:test"), "file")

    def test_icon_name_artifact(self) -> None:
        """Test artifact_icon_name tag on artifacts."""
        for k, v in ARTIFACT_CATEGORY_ICON_NAMES.items():
            artifact = Artifact(category=k)
            self.assertEqual(artifact_icon_name(artifact), v)
        artifact = Artifact(category="debusine:test")
        self.assertEqual(artifact_icon_name(artifact), "file")

    def test_artifact_link(self) -> None:
        artifact, _ = self.playground.create_artifact()

        self.assertEqual(
            artifact_link(artifact.id),
            f'<a href="{artifact.get_absolute_url()}">'
            f'{artifact.get_label()}</a>',
        )

    def test_artifact_link_deleted(self) -> None:
        artifact, _ = self.playground.create_artifact()
        non_existing_id = artifact.id
        artifact.delete()

        self.assertEqual(
            artifact_link(non_existing_id), f"{non_existing_id} (deleted)"
        )

    def test_artifact_links(self) -> None:
        artifact, _ = self.playground.create_artifact()
        artifact2, _ = self.playground.create_artifact()
        artifact2.delete()
        links = artifact_links([artifact.id, artifact2.id])

        self.assertEqual(
            links,
            f"{artifact_link(artifact.id)}, {artifact_link(artifact2.id)}",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for scheduler tags backend functions."""

from typing import override
from unittest import mock

import debusine.worker.tags as wtags
from debusine.server.tags import Tagset, make_worker_provided_tagset
from debusine.test import TestCase
from debusine.utils.tag_utils import (
    CandidateTags,
    DerivationRules,
    ProvenanceRestrictions,
)


class TagsetTests(TestCase):
    """Tests for :py:class:`Tagset`."""

    def test_init(self) -> None:
        t = Tagset()
        self.assertEqual(t.restrictions, [])
        self.assertEqual(t.derivation_rules, [])
        self.assertEqual(t.tags, set())
        self.assertEqual(t.log, [])

    def test_plain(self) -> None:
        t = Tagset()
        t.add("test", None)
        t.add("test", ["foo"])
        t.add("test", {"bar", "baz"})
        t.finalize()
        self.assertEqual(t.tags, {"foo", "bar", "baz"})

    def test_add_candidates(self) -> None:
        c: CandidateTags[str] = CandidateTags()
        c.add("test", {"foo"})
        c.add("test1", {"bar", "baz"})

        t = Tagset()
        with self.spy_object(t, "add") as add:
            t.add_candidates(c)

        # Ensure that adding candidates is the same as calling add with their
        # contents
        add.assert_has_calls(
            [
                mock.call("test", {"foo"}),
                mock.call("test1", {"bar", "baz"}),
            ]
        )

    def test_restrictions(self) -> None:
        restrictions = ProvenanceRestrictions("test")
        restrictions.add_exact("group", ["system"])
        restrictions.add_exact("test", ["system"])
        t = Tagset()
        t.restrictions.append(restrictions)
        t.add("system", {"test", "foo"})
        t.add("workspace", {"group", "bar"})
        t.finalize()
        self.assertEqual(t.tags, {"test", "foo", "bar"})
        self.assertEqual(
            t.log, ["test provenance restrictions forbids tags group"]
        )

    def test_derivation(self) -> None:
        derivation = DerivationRules(provenance="system")
        derivation.add_rule("dm or dd", {"debian"})
        t = Tagset()
        t.derivation_rules.append(derivation)
        t.add("system", {"dm", "foo"})
        t.finalize()
        self.assertEqual(t.tags, {"dm", "debian", "foo"})
        self.assertEqual(t.log, ["system derivation rules adds tags debian"])

    def test_derivation_exists_but_does_not_match(self) -> None:
        derivation = DerivationRules(provenance="system")
        derivation.add_rule("dm or dd", {"debian"})
        t = Tagset()
        t.derivation_rules.append(derivation)
        t.add("system", {"foo"})
        t.finalize()
        self.assertEqual(t.tags, {"foo"})
        self.assertEqual(t.log, [])

    def test_derivation_restrictions(self) -> None:
        restrictions = ProvenanceRestrictions("test")
        restrictions.add_exact("group", ["system"])
        derivation = DerivationRules(provenance="workspace")
        derivation.add_rule("dm or dd", {"debian"})
        derivation.add_rule("foo", {"group"})
        t = Tagset()
        t.restrictions.append(restrictions)
        t.derivation_rules.append(derivation)
        t.add("system", {"test", "foo", "dm"})
        t.finalize()
        self.assertEqual(t.tags, {"test", "foo", "dm", "debian"})
        self.assertEqual(
            t.log,
            [
                "workspace derivation rules adds tags debian, group",
                "test provenance restrictions forbids tags group",
            ],
        )


class TestWorkerProvidedTags(TestCase):
    """Test worker provided tags."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.tagset = make_worker_provided_tagset()

    def test_class(self) -> None:
        self.tagset.add(
            wtags.ProvenanceProvided.WORKER, [wtags.CLASS_PREFIX + "worker"]
        )
        self.tagset.add(
            wtags.ProvenanceProvided.ADMIN, [wtags.CLASS_PREFIX + "admin"]
        )
        self.tagset.add(
            wtags.ProvenanceProvided.SYSTEM, [wtags.CLASS_PREFIX + "system"]
        )
        self.tagset.finalize()
        self.assertEqual(self.tagset.tags, {wtags.CLASS_PREFIX + "admin"})



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Helpers for testing permissions."""

import itertools
from collections.abc import Iterable
from typing import cast

from debusine.db.models.permissions import RoleBase, Roles, WithImplies
from debusine.test.django import TestCase


class RolesTestCase[R: RoleBase](TestCase):
    """Helper functions for testing roles enums."""

    roles_class: type[R]

    @property
    def entries(self) -> list[RoleBase]:
        """List all roles defined in the class."""
        return list(cast(Iterable[R], self.roles_class))

    def assertPartialOrdering(self) -> None:
        """Ensure implications form a partial ordering."""
        # Ensure reflexive
        for a in self.entries:
            with self.subTest(a=repr(a)):
                assert isinstance(a, WithImplies)
                self.assertTrue(a.implies(a))

        if len(self.entries) <= 1:  # pragma: no cover
            return

        # Ensure antisymmetric
        for a, b in itertools.product(self.entries, self.entries):
            with self.subTest(a=repr(a), b=repr(b)):
                assert isinstance(a, WithImplies)
                assert isinstance(b, WithImplies)
                if a.implies(b) and b.implies(a):
                    self.assertEqual(a, b)

        # Ensure transitive
        for a, b, c in itertools.product(
            self.entries, self.entries, self.entries
        ):
            with self.subTest(a=repr(a), b=repr(b), c=repr(c)):
                assert isinstance(a, WithImplies)
                assert isinstance(b, WithImplies)
                assert isinstance(c, WithImplies)
                if a.implies(b) and b.implies(c):
                    self.assertTrue(a.implies(c))

    def assertRolesInvariants(self) -> None:
        """Check invariants in Roles enum."""
        self.assertPartialOrdering()

    def assertChoices(self) -> None:
        """Ensure choices returns the right values."""
        assert issubclass(self.roles_class, Roles)
        self.assertEqual(
            self.roles_class.choices, [(r.value, r.label) for r in self.entries]
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command manage_channel_notification."""

import copy
import io
from typing import Any, ClassVar, override

import yaml
from django.core.management import CommandError

from debusine.db.models import NotificationChannel
from debusine.django.management.tests import call_command
from debusine.test.django import TestCase


class ManageNotificationChannelTests(TestCase):
    """Tests for the manage_notification_channel command."""

    email_data: ClassVar[dict[str, Any]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.email_data = {
            "from": "debusine@debusine.com",
            "to": ["user@example.com"],
        }

    def test_change_name(self) -> None:
        """Verify change name of a notification channel."""
        old_name = "lts"
        new_name = "new-lts"

        NotificationChannel.objects.create(
            name="lts",
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        stdout, stderr, exit_code = call_command(
            "manage_notification_channel",
            "change-name",
            old_name,
            new_name,
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertTrue(
            NotificationChannel.objects.filter(name=new_name).exists()
        )

    def test_change_data_from_file(self) -> None:
        """Verify change data of a notification channel from file."""
        new_data = copy.deepcopy(self.email_data)
        new_data["from"] = "a-new-email@example.com"

        name = "lts"

        NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        new_data_file = self.create_temporary_file(
            contents=yaml.safe_dump(new_data).encode("utf-8")
        )

        stdout, stderr, exit_code = call_command(
            "manage_notification_channel",
            "change-data",
            name,
            "--data",
            str(new_data_file),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(
            NotificationChannel.objects.get(name=name).data, new_data
        )

    def test_change_data_from_stdin(self) -> None:
        """Verify change data of a notification channel from stdin."""
        new_data = copy.deepcopy(self.email_data)
        new_data["from"] = "a-new-email@example.com"

        name = "lts"

        NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        stdout, stderr, exit_code = call_command(
            "manage_notification_channel",
            "change-data",
            name,
            stdin=io.StringIO(yaml.safe_dump(new_data)),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(
            NotificationChannel.objects.get(name=name).data, new_data
        )

    def test_change_data_invalid_data(self) -> None:
        """Command raise CommandError: invalid data."""
        name = "deblts-email"
        NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        new_data_file = self.create_temporary_file(
            contents=yaml.safe_dump({"invalid": "data"}).encode("utf-8")
        )

        with self.assertRaisesRegex(CommandError, "^Invalid data") as exc:
            call_command(
                "manage_notification_channel",
                "change-data",
                name,
                "--data",
                str(new_data_file),
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_change_name_name_does_not_exist(self) -> None:
        """Name cannot be changed: NotificationChannel does not exist."""
        name = "does-not-exist"
        expected_error = f'NotificationChannel "{name}" does not exist'

        with self.assertRaisesMessage(CommandError, expected_error) as exc:
            call_command(
                "manage_notification_channel",
                "change-name",
                name,
                "new-name",
            )

        self.assertEqual(exc.exception.returncode, 3)



# Generated by Django 4.2.23 on 2026-01-08 10:36

import django.db.models.deletion
import django.db.models.functions.comparison
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0040_workrequest_pending_at'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='group',
            name='db_group_unique_name_scope',
        ),
        migrations.AddField(
            model_name='group',
            name='workspace',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='db.workspace',
            ),
        ),
        migrations.AddConstraint(
            model_name='group',
            constraint=models.UniqueConstraint(
                models.F('name'),
                models.F('scope'),
                django.db.models.functions.comparison.Coalesce(
                    'workspace', models.Value(-1)
                ),
                name='db_group_unique_name_scope_workspace',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface, class-based subparsers."""

import abc
import argparse
import contextlib
import copy
import http
import inspect
import logging
import math
import os
import sys
from collections import defaultdict
from collections.abc import Generator, Iterable
from contextlib import contextmanager
from pathlib import Path
from typing import Any, ClassVar, NoReturn, Optional, TextIO, override

import rich.console
import rich.json
import rich.text
import rich.theme
import yaml

from debusine.client import exceptions
from debusine.client.client_utils import copy_file
from debusine.client.config import ConfigHandler
from debusine.client.debusine import Debusine
from debusine.client.exceptions import DebusineError
from debusine.client.models import FileResponse, StrictBaseModel


def _get_help(obj: Any) -> str | None:
    """Extract the first line from an object's docstring."""
    if obj.__doc__ is None:
        raise AssertionError(f"{obj!r} lacks a docstring")
    assert isinstance(obj.__doc__, str)
    return " ".join(
        stripped
        for line in obj.__doc__.splitlines()
        if (stripped := line.strip())
    ).rstrip(".")


class Command(abc.ABC):
    """
    Base class for defining argparse subparsers as classes.

    Subcommands are self-registering and self-documenting:

    * The subcommand name is the lowercased class name by default, or the
      ``name`` attribute of the class if present.
    * The help string is the first line of the class docstring.
    * ``Command.add_subparsers`` adds all registered subparsers to an existing
      ``ArgumentParser.add_subparsers`` instance
    """

    #: Command name
    name: ClassVar[str]
    #: String to show on help if the command is deprecated
    deprecated: ClassVar[str | None]
    #: Registry of commands by command group. A key of None means "no group"
    commands_by_group: dict[str | None, list[type["Command"]]] = defaultdict(
        list
    )

    #: Set to True if YAML data was provided on stdin
    yaml_in_input: bool
    #: Rich console, or None if --yaml is in use
    console: rich.console.Console | None

    @override
    def __init_subclass__(
        cls,
        name: str | None = None,
        group: str | None = None,
        deprecated: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Register this class as a subcommand."""
        super().__init_subclass__(**kwargs)
        if inspect.isabstract(cls):
            return
        cls.name = name or cls.__name__.lower()
        cls.deprecated = deprecated
        cls.commands_by_group[group].append(cls)

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Make the parsed arguments available as ``self.args``."""
        self.argument_parser = parser
        self.args = args
        self.yaml_in_input = False
        if self.args.yaml:
            self.console = None
        else:
            self.console = self.make_console()

    @classmethod
    def add_subparser(
        cls,
        subparsers: "argparse._SubParsersAction[Any]",
        group: str,
        *,
        argument: str | None = None,
        argument_help: str,
        subparser_help: str,
    ) -> None:
        """
        Add a subparser with entries from a Command group.

        :param group: the group name
        :param argument: the subcommand name for the parser (defaults to the
          group name)
        :param argument_help: help string for the subcommand
        :param subparser_help: help string for the subparser
        """
        parser = subparsers.add_parser(argument or group, help=argument_help)
        parser_commands = parser.add_subparsers(
            help=subparser_help, required=True
        )
        cls.add_parsers(parser_commands, group)

    @classmethod
    def add_parsers(
        cls,
        subparsers: "argparse._SubParsersAction[Any]",
        group: str | None = None,
    ) -> None:
        """Add all registered Command subclasses to the subparsers instance."""
        if group not in cls.commands_by_group:
            raise KeyError(
                f"Command group {group!r} not found in "
                + ', '.join(repr(x) for x in cls.commands_by_group)
            )
        for command_cls in cls.commands_by_group[group]:
            # Skip deprecated commands during shell completion
            if command_cls.deprecated and "_ARGCOMPLETE" in os.environ:
                continue

            description = _get_help(command_cls)
            kwargs: dict[str, Any] = {"help": description}
            if command_cls.deprecated:
                del kwargs["help"]
                description = f"Deprecated: {command_cls.deprecated}"
            subparser = subparsers.add_parser(
                command_cls.name, description=description, **kwargs
            )
            subparser.set_defaults(command_cls=command_cls)
            command_cls.configure(subparser)

    @classmethod
    def create(
        cls, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> Optional["Command"]:
        """
        Run the command activated by the parsed command line.

        :param args: the parsed command line arguments
        :returns: True if a command was selected and run, False otherwise
        """
        if (command_cls := getattr(args, "command_cls", None)) is not None:
            assert isinstance(command_cls, type)
            assert issubclass(command_cls, Command)
            return command_cls(parser, args)
        return None

    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        """Configure the ArgumentParser for this subcommand."""
        parser.add_argument(
            "--yaml",
            action="store_true",
            help="Output data as machine readable YAML.",
        )

    @abc.abstractmethod
    def run(self) -> None:
        """Run the command."""

    def run_porcelain(self) -> None:
        """Run the command, managing well known exceptions."""
        try:
            self.run()
        except DebusineError as e:
            self.show_error(e)

    def make_console(self) -> rich.console.Console:
        """Create a Rich console to use for output."""
        theme = rich.theme.Theme({"feedback": "bold"})
        return rich.console.Console(theme=theme)

    @staticmethod
    def _fail(error: object, *, summary: str | None = None) -> NoReturn:
        print(error, file=sys.stderr)
        if summary is not None:
            print(summary, file=sys.stderr)
        raise SystemExit(3)

    @staticmethod
    def _print_yaml(data: Any, *, file: TextIO | None = None) -> None:
        """Print data to stdout as yaml."""
        output = yaml.safe_dump(data, sort_keys=False, width=math.inf)

        print(output, file=file, end="")

    def print_yaml_output(self, data: Any) -> None:
        """Print the command output as a data structure in YAML."""
        if self.yaml_in_input:
            print("---")  # avoid confusion between input and output
        self._print_yaml(data)

    def _show_error_yaml(self, error: Exception) -> None:
        """Output a Debusine error as YAML."""
        if isinstance(error, DebusineError):
            self.print_yaml_output(
                {"result": "failure", "error": error.asdict()}
            )
        else:
            self.print_yaml_output({"result": "failure", "error": str(error)})

    def _show_error_rich(self, error: Exception) -> None:
        """Output a Debusine error for humans."""
        assert self.console is not None
        if isinstance(error, DebusineError):
            if error.status_code is not None:
                self.console.print(
                    f"Error ({error.status_code}):",
                    style="red",
                    markup=False,
                    highlight=False,
                    end=" ",
                )
            else:
                self.console.print(
                    "Error:",
                    style="red",
                    markup=False,
                    highlight=False,
                    end=" ",
                )
            self.console.print(
                error.title, style="red", markup=False, highlight=False
            )
            if error.detail is not None:
                self.console.print(
                    error.detail, style="yellow", markup=False, highlight=False
                )
            if error.validation_errors is not None:
                self.console.print(
                    rich.json.JSON.from_data(error.validation_errors, indent=2)
                )
        else:
            self.console.print("[red]Command failed:[/]", end=" ")
            self.console.print(str(error), highlight=False, markup=False)

    def show_error(self, error: Exception) -> NoReturn:
        """Output a Debusine error."""
        if self.args.yaml:
            self._show_error_yaml(error)
        else:
            self._show_error_rich(error)
        raise SystemExit(3)

    def feedback(
        self,
        text: str,
        style: str | rich.style.Style | None = "feedback",
        markup: bool | None = None,
        highlight: bool | None = None,
    ) -> None:
        """Show feedback to the user."""
        if self.console:
            self.console.print(
                text, style=style, markup=markup, highlight=highlight
            )
        else:
            # Strip Rich markup
            rich_text = rich.text.Text.from_markup(text)
            plain_text = rich_text.plain
            logging.getLogger("debusine").info("%s", plain_text)

    @classmethod
    def _parse_yaml_data(cls, data_yaml: str) -> dict[str, Any]:
        if data_yaml.strip() == "":
            cls._fail("Error: data must be a dictionary. It is empty")
        try:
            data = yaml.safe_load(data_yaml)
        except yaml.YAMLError as err:
            cls._fail(
                f"Error parsing YAML: {err}",
                summary="Fix the YAML data",
            )

        if (data_type := type(data)) is not dict:
            cls._fail(
                f"Error: data must be a dictionary. It is: {data_type.__name__}"
            )

        assert isinstance(data, dict)
        return data

    @staticmethod
    @contextlib.contextmanager
    def preserve_registry() -> Generator[None]:
        """Make Command registry changes ephemeral."""
        orig = Command.commands_by_group
        Command.commands_by_group = copy.deepcopy(orig)
        try:
            yield
        finally:
            Command.commands_by_group = orig


class DebusineCommand(Command, abc.ABC):
    """Command that uses a debusine object."""

    debusine: Debusine

    def _build_debusine_object(self) -> Debusine:
        """Return the debusine object matching the command line parameters."""
        return Debusine(
            base_api_url=self.server_configuration.api_url,
            scope=self.args.scope or self.server_configuration.scope,
            api_token=self.server_configuration.api_token,
            logger=logging.getLogger("debusine"),
        )

    def _setup_logging(self) -> None:
        logging_level = logging.WARNING if self.args.silent else logging.INFO

        logger = logging.getLogger("debusine")
        logger.propagate = False
        logger.setLevel(logging_level)

        handler = logging.StreamHandler(sys.stderr)
        formatter = logging.Formatter('%(message)s')
        handler.setFormatter(formatter)

        logger.addHandler(handler)

    def _setup_http_logging(self) -> None:
        if self.args.debug:
            """Do setup urllib3&requests traces."""
            # https://docs.python-requests.org/en/latest/api/#api-changes
            # Debug connection establishment
            # Use a separate DEBUG logger
            urllib3_log = logging.getLogger("urllib3")
            urllib3_log.setLevel(logging.DEBUG)
            sh = logging.StreamHandler()
            sh.setLevel(logging.DEBUG)
            sh.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
            urllib3_log.addHandler(sh)

            # Debug at http.client level (requests->urllib3->http.client)
            # Use a separate DEBUG logger
            # Display the REQUEST, including HEADERS and DATA, and
            # RESPONSE with HEADERS but without DATA.  The only thing
            # missing will be the response.body which is not logged.
            http.client.HTTPConnection.debuglevel = 1
            # Divert to new logger (stderr) to avoid polluting CLI output
            requests_log = logging.getLogger("requests")
            requests_log.setLevel(logging.DEBUG)
            requests_log.addHandler(sh)

            def print_to_log(*args: str) -> None:
                requests_log.debug(" ".join(args))

            setattr(http.client, "print", print_to_log)

    @contextmanager
    def _api_call_or_fail(self) -> Generator[None]:
        """
        Context manager to handle failures from calling a method.

        :raises: exceptions.NotFoundError: server returned 404.
        :raises: UnexpectedResponseError: e.g. invalid JSON.
        :raises: ClientConnectionError: e.g. cannot connect to the server.
        :raises: DebusineError (via method() call) when debusine server
          reports an error.
        """
        try:
            yield
        except exceptions.NotFoundError as exc:
            self._fail(exc)
        except exceptions.UnexpectedResponseError as exc:
            self._fail(exc)
        except exceptions.ClientForbiddenError as server_error:
            self._fail(f'Server rejected connection: {server_error}')
        except exceptions.ClientConnectionError as client_error:
            self._fail(f'Error connecting to debusine: {client_error}')

    def _fetch_local_file(
        self, src: Path, dest: Path, artifact_file: FileResponse
    ) -> None:
        """Copy src to dest and verify that its hash matches artifact_file."""
        if not src.exists():
            self._fail(f"{str(src)!r} does not exist.")
        hashes = copy_file(src, dest, artifact_file.checksums.keys())
        if hashes["size"] != artifact_file.size:
            self._fail(
                f"{str(src)!r} size mismatch (expected {artifact_file.size} "
                f"bytes)"
            )

        for hash_name, expected_value in artifact_file.checksums.items():
            if hashes[hash_name] != expected_value:
                self._fail(
                    f"{str(src)!r} hash mismatch (expected {hash_name} "
                    f"= {expected_value})"
                )

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Make the parsed arguments available as ``self.args``."""
        super().__init__(parser, args)
        server_name: str | None = os.environ.get("DEBUSINE_SERVER_NAME")
        if self.args.server is not None:
            server_name = self.args.server

        configuration = ConfigHandler(
            server_name=server_name, config_file_path=self.args.config_file
        )

        try:
            self.server_configuration = configuration.server_configuration()
        except ValueError as exc:
            self._fail(exc)

        self._setup_logging()
        self.debusine = self._build_debusine_object()
        self._setup_http_logging()

    @classmethod
    def print_server_error(cls, error: DebusineError) -> None:
        """Print a server error in YAML."""
        description: dict[str, Any] = {
            "result": "failure",
            "error": error.asdict(),
        }
        cls._print_yaml(description, file=sys.stderr)


class WorkspaceCommand(DebusineCommand, abc.ABC):
    """Command that operates on a workspace."""

    workspace: str
    workspace_arg_set: bool

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--workspace",
            "-w",
            type=str,
            metavar="NAME",
            help="Workspace name",
        )

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Populate self.workspace."""
        super().__init__(parser, args)
        if args.workspace is None:
            configured_default = self.server_configuration.default_workspace
            self.workspace = configured_default or "System"
            self.workspace_arg_set = False
        else:
            self.workspace = args.workspace
            self.workspace_arg_set = True


class ModelCommand[Model: StrictBaseModel](Command, abc.ABC):
    """Base for commands that list and show client models."""

    def _list_yaml(self, instances: Iterable[Model]) -> None:
        """Print the list as YAML."""
        if self.yaml_in_input:
            print("---")  # avoid confusion between input and output
        self._print_yaml([i.model_dump(mode="json") for i in instances])

    @abc.abstractmethod
    def _list_rich(self, instances: Iterable[Model]) -> None:
        """Print the list as a table."""

    def list(self, instances: Iterable[Model]) -> None:
        """Output a list of instances."""
        if self.args.yaml:
            self._list_yaml(instances)
        else:
            self._list_rich(instances)

    def _show_yaml(self, instance: Model) -> None:
        """Show an instance in machine parsable YAML."""
        if self.yaml_in_input:
            print("---")  # avoid confusion between input and output
        self._print_yaml(instance.model_dump(mode="json"))

    @abc.abstractmethod
    def _show_rich(self, instance: Model) -> None:
        """Show an instance for humans."""

    def show(self, instance: Model) -> None:
        """Output a list of collections."""
        if self.args.yaml:
            self._show_yaml(instance)
        else:
            self._show_rich(instance)


class InputDataCommand(Command, abc.ABC):
    """Common code to read input for ``--data``."""

    def read_input_data(self, file: str) -> dict[str, Any]:
        """
        Read input data given a file name.

        If file is ``"-"``, read from stdin.
        """
        data: str
        match file:
            case "-":
                if sys.stdout.isatty() and sys.stdin.isatty():
                    console = self.console or self.make_console()
                    console.print(
                        "[bold]Waiting for YAML input...[/]"
                        " [grey50]([bold]CTRL+D[/bold] to terminate the input."
                        " Empty input defaults to [green]'{}'[/green])[/]",
                        highlight=False,
                    )
                data = sys.stdin.read()
                sys.stdin.close()
                self.yaml_in_input = True
                # Empty stdin defaults to {}
                if not data or data.isspace():
                    return {}
            case _:
                try:
                    with open(file) as fd:
                        data = fd.read()
                except OSError as e:
                    print(f"can't open {file!r}: {e}", file=sys.stderr)
                    raise SystemExit(2)

        return self._parse_yaml_data(data)


class OptionalInputDataCommand(InputDataCommand, abc.ABC):
    """Command reading YAML in input with ``--data``, if provided."""

    #: Parsed YAML data
    input_data: dict[str, Any] | None

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Read YAML data from input."""
        super().__init__(parser, args)
        if self.args.data is None:
            self.input_data = None
        else:
            self.input_data = self.read_input_data(self.args.data)

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--data",
            type=str,
            default=None,
            help="File path (or - for stdin) to read the data "
            "for the asset. YAML format. Default: data is an empty dict.",
        )


class RequiredInputDataCommand(InputDataCommand, abc.ABC):
    """Command reading YAML in input with ``--data``."""

    #: Parsed YAML data
    input_data: dict[str, Any]

    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        """Read YAML data from input."""
        super().__init__(parser, args)
        self.input_data = self.read_input_data(self.args.data)

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--data",
            type=str,
            default="-",
            help="File path (or - for stdin) to read the data "
            "for the asset. YAML format. Default: read from stdin.",
        )
